# Copyright 2017 MDSLAB - University of Messina
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from iotronic_lightningrod.plugins import Plugin
from iotronic_lightningrod.plugins import pluginApis as API

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

# User imports
device = API.getBoardGpio()


class Worker(Plugin.Plugin):

    def __init__(self, uuid, name, q_result, params=None):
        super(Worker, self).__init__(uuid, name, q_result, params)

    def run(self):
        LOG.info("Input parameters: " + str(self.params))

        device._setGPIOs("D13", "out", str(self.params['pin']))

        LOG.info("Plugin " + self.name + " process completed!")
        self.q_result.put("Led D13: " + str(self.params['pin']))



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import os
import time

from iotronic_lightningrod.devices.gpio import Gpio

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

i2c_path = "/sys/devices/mcuio/0:0.0/0:1.4/i2c-0"
device1_path = i2c_path + "/0-0060/iio:device1/"
device0_path = "/sys/bus/iio/devices/iio:device0/"


class YunGpio(Gpio.Gpio):

    def __init__(self):
        super(YunGpio, self).__init__("yun")

        self.MAPPING = {
            'D8': '104',
            'D9': '105',
            'D10': '106',
            'D11': '107',
            'D5': '114',
            'D13': '115',
            'D3': '116',
            'D2': '117',
            'D4': '120',
            'D12': '122',
            'D6': '123'}

        # LOG.info("Arduino YUN gpio module importing...")

    def EnableGPIO(self):
        """Enable GPIO (device0).

        """
        try:

            with open('/sys/bus/iio/devices/iio:device0/enable', 'a') as f:
                f.write('1')

            result = "  - GPIO enabled!\n"
            LOG.info(result)

        except Exception as err:
            LOG.error("Error enabling GPIO (device0): " + str(err))

    def DisableGPIO(self):
        """Disable GPIO (device0).

        """
        try:
            with open('/sys/bus/iio/devices/iio:device0/enable', 'a') as f:
                f.write('0')

            result = "  - GPIO disabled!\n"
            LOG.info(result)
        except Exception as err:
            LOG.error("Error disabling GPIO (device0): " + str(err))

    def EnableI2c(self):
        """Enable i2c device (device1).

        From ideino-linino-lib library:
            Board.prototype.addI2c = function(name, driver, addr, bus)
                board.addI2c('BAR', 'mpl3115', '0x60', 0):
                - i2c_device.driver: mpl3115
                - i2c_device.addr: 0x60
                - i2c_device.name: BAR
                - i2c_device.bus: 0

        """

        try:

            if os.path.exists('/sys/bus/i2c/devices/i2c-0/0-0060'):
                result = "  - I2C device already enabled!"

            else:

                with open('/sys/bus/i2c/devices/i2c-0/new_device', 'a') as f:
                    # 'echo '+i2c_device.driver+' '+i2c_device.addr+ '
                    f.write('mpl3115 0x60')
                    result = "  - I2C device enabled!"

            LOG.info(result)

        except Exception as err:
            LOG.error("Error enabling I2C (device1): " + str(err))

    def i2cRead(self, sensor):
        """Read i2c raw value.

        sensor options:
        - in_pressure_raw
        - in_temp_raw

        :param sensor: name of the sensor connected to I2C port
        :return: I2C raw value

        """
        try:

            with open(device1_path + "in_" + sensor + "_raw") as raw:
                value = raw.read()

        except Exception as err:
            LOG.error("Error reading I2C device: " + str(err))
            value = None

        return value

    def setPIN(self, DPIN, value):
        """Function to set digital PIN value.

        :param DPIN: pin
        :param value: value to set the pin

        """
        try:
            with open('/sys/class/gpio/' + DPIN + '/value', 'a') as f:
                f.write(value)

        except Exception as err:
            LOG.error("Error setting PIN value: " + str(err))

    def _setGPIOs(self, Dpin, direction, value):
        """GPIO mapping on lininoIO

            -------------------------
            GPIO n.     OUTPUT
            104	        D8
            105	        D9
            106	        D10
            107	        D11
            114	        D5
            115	        D13
            116	        D3
            117	        D2
            120	        D4
            122	        D12
            123	        D6

        """

        try:

            with open('/sys/class/gpio/export', 'a') as f_export:
                f_export.write(self.MAPPING[Dpin])

            with open('/sys/class/gpio/' + Dpin + '/direction', 'a') as f_dir:
                f_dir.write(direction)

            with open('/sys/class/gpio/' + Dpin + '/value', 'a') as f_value:
                f_value.write(value)

            with open('/sys/class/gpio/' + Dpin + '/value') as f_value:
                result = "PIN " + Dpin + " value " + f_value.read()

        except Exception as err:
            LOG.error("Error setting GPIO value: " + str(err))
            result = None

        return result

    def _readVoltage(self, pin):

        try:
            with open(device0_path + "in_voltage_" + pin + "_raw") as raw:
                voltage = raw.read()
                # print("VOLTAGE: " + voltage)

        except Exception as err:
            LOG.error("Error reading voltage: " + str(err))
            voltage = None

        return voltage

    def blinkLed(self):
        """LED: 13. There is a built-in LED connected to digital pin 13.

        When the pin has HIGH value, the LED is on,
        when the pin has LOW value, it is off.

        """
        with open('/sys/class/gpio/export', 'a') as f:
            f.write('115')

        with open('/sys/class/gpio/D13/direction', 'a') as f:
            f.write('out')

        with open('/sys/class/gpio/D13/value', 'a') as f:
            f.write('1')

        time.sleep(2)

        with open('/sys/class/gpio/D13/value', 'a') as f:
            f.write('0')



# Copyright 2017 MDSLAB - University of Messina
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from iotronic_lightningrod.devices.gpio import yun
from iotronic_lightningrod.plugins import Plugin

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

# User imports
import datetime
import math
import time

ADCres = 1023.0
Beta = 3950
Kelvin = 273.15
Rb = 10000
Ginf = 120.6685

# User global variables
resource_id = ""  # temperature resource id
action_URL = "http://smartme-data.unime.it/api/3/action/datastore_upsert"
api_key = ''
headers = {
    "Content-Type": "application/json",
    'Authorization': "" + api_key + ""
}
polling_time = 10


class Worker(Plugin.Plugin):
    def __init__(self, name, params=None):
        super(Worker, self).__init__(name, params)

    def run(self):

        device = yun.YunGpio()

        while (self._is_running):

            voltage = device._readVoltage("A0")

            Rthermistor = float(Rb) * (float(ADCres) / float(voltage) - 1)

            rel_temp = float(Beta) / (math.log(
                float(Rthermistor) * float(Ginf))
            )
            temp = rel_temp - Kelvin

            m_value = str(temp)
            m_timestamp = datetime.datetime.now().strftime(
                '%Y-%m-%dT%H:%M:%S.%f'
            )

            LOG.info(m_value + " - " + m_timestamp)

            time.sleep(polling_time)



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import _pickle as pickle

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


class ObjectSerializer(object):
    """A PluginObject-aware Serializer.

    This implements the Oslo Serializer interface and provides the
    ability to serialize and deserialize PluginObject entities.
    Any service that needs to accept or return PluginObject as
    arguments or result values should pass this to its RpcProxy
    and RpcDispatcher objects.
    """

    def serialize_entity(self, entity):

        dumped = pickle.dumps(entity, 0)

        # LOG.debug(" - plugin serialized")

        return dumped

    def deserialize_entity(self, entity):

        loaded = pickle.loads(str.encode(entity))

        # LOG.debug(" - plugin deserialized")

        return loaded



#    Copyright 2017 MDSLAB - University of Messina All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

# Autobahn imports
from autobahn.asyncio.component import Component
from autobahn.wamp import exception

# OSLO imports
from oslo_config import cfg
from oslo_log import log as logging

# MODULES imports
import asyncio
import inspect
import os
import pkg_resources
import psutil
import signal
import ssl
from stevedore import extension
import sys
import txaio

from pip._vendor import pkg_resources

# IoTronic imports
from iotronic_lightningrod.Board import Board
from iotronic_lightningrod.common.exception import timeoutALIVE
from iotronic_lightningrod.common.exception import timeoutRPC
from iotronic_lightningrod.common import utils
from iotronic_lightningrod.common.utils import get_version
import iotronic_lightningrod.wampmessage as WM


# Global variables
LOG = logging.getLogger(__name__)


lr_opts = [
    cfg.StrOpt('lightningrod_home',
               default='/var/lib/iotronic',
               help=('Lightning-rod Home Data')),
    cfg.BoolOpt('skip_cert_verify',
                default=True,
                help=('Flag for skipping the verification of the server cert '
                      '(for the auto-signed ones)')),
    cfg.IntOpt('connection_timer',
               default=10,
               help=('IoTronic connection RPC timer')),
    cfg.IntOpt('alive_timer',
               default=600,
               help=('Wamp websocket check time')),
    cfg.IntOpt('rpc_alive_timer',
               default=3,
               help=('RPC alive response time threshold')),
]

CONF = cfg.CONF
CONF.register_opts(lr_opts)

SESSION = None
global board
board = None
reconnection = False
RPC = {}
RPC_devices = {}
RPC_proxies = {}
zombie_alert = True

# ASYNCIO
loop = None
component = None
txaio.start_logging(level="info")
RUNNER = None
connected = False

global MODULES
MODULES = {}


class LightningRod(object):

    def __init__(self):

        LogoLR()

        LOG.info(' - version: ' +
                 str(get_version("iotronic-lightningrod")))
        LOG.info(' - PID: ' + str(os.getpid()))

        LOG.info("LR available modules: ")
        for ep in pkg_resources.iter_entry_points(group='s4t.modules'):
            LOG.info(" - " + str(ep))

        logging.register_options(CONF)
        DOMAIN = "s4t-lightning-rod"
        CONF(project='iotronic')
        logging.setup(CONF, DOMAIN)

        if (utils.checkIotronicConf(CONF)):

            if CONF.debug:
                txaio.start_logging(level="debug")

            signal.signal(signal.SIGINT, self.stop_handler)

            LogoLR()

            LOG.info('Lightning-rod: ')
            LOG.info(' - version: ' +
                     str(get_version("iotronic-lightningrod")))
            LOG.info(' - PID: ' + str(os.getpid()))
            LOG.info(' - Logs: ' + CONF.log_file)
            LOG.info(" - Home: " + CONF.lightningrod_home)
            LOG.info(" - Alive Check timer: " + str(CONF.alive_timer) +
                     " seconds")
            LOG.info(" - RPC-Alive Check timer: " + str(CONF.rpc_alive_timer) +
                     " seconds")

            global board
            board = Board()

            self.w = WampManager(board.wamp_config)

            self.w.start()

        else:
            Bye()

    def stop_handler(self, signum, frame):

        try:
            # No zombie alert activation
            zombie_alert = False
            LOG.info("LR is shutting down...")
            self.w.stop()
            Bye()
        except Exception as e:
            LOG.error("Error closing LR")


class WampManager(object):
    """WAMP Manager: through this LR manages the connection to Crossbar server.

    """

    def __init__(self, wamp_conf):

        # wampConnect configures and manages the connection to Crossbar server.
        wampConnect(wamp_conf)

    def start(self):
        LOG.info(" - starting Lightning-rod WAMP server...")

        global loop
        loop = asyncio.get_event_loop()
        component.start(loop)
        loop.run_forever()

    def stop(self):
        LOG.info("Stopping WAMP agent server...")
        # Canceling pending tasks and stopping the loop
        asyncio.gather(*asyncio.Task.all_tasks()).cancel()
        LOG.info("WAMP server stopped!")


async def wamp_checks(session):

    while (True):

        try:

            # LOG.debug("ALIVE sending...")

            with timeoutALIVE(seconds=CONF.rpc_alive_timer, action="ws_alive"):
                res = await session.call(
                    str(board.agent) + u'.stack4things.wamp_alive',
                    board_uuid=board.uuid,
                    board_name=board.name
                )

            LOG.debug("WampCheck attempt " + str(res))

        except exception.ApplicationError as e:
            LOG.error(" - Iotronic Connection RPC error: " + str(e))
            # Iotronic is offline the board can not call
            # the "stack4things.alive" RPC.
            # The board will disconnect from WAMP agent and retry later.
            global reconnection
            reconnection = True
            utils.destroyWampSocket()

        try:
            await asyncio.sleep(CONF.alive_timer)
        except Exception as e:
            LOG.warning(" - asyncio alert: " + str(e))


async def IotronicLogin(board, session, details):
    """Function called to connect the board to Iotronic.

    The board:
     1. logs in to Iotronic
     2. loads the modules

    :param board:
    :param session:
    :param details:

    """

    LOG.info("IoTronic Authentication:")

    global reconnection

    try:

        rpc = str(board.agent) + u'.stack4things.connection'

        with timeoutRPC(seconds=CONF.connection_timer, action=rpc):
            res = await session.call(
                rpc,
                uuid=board.uuid,
                session=details.session
            )

            w_msg = WM.deserialize(res)

            if w_msg.result == WM.SUCCESS:

                LOG.info(" - Access granted to Iotronic.")

                # WS ALIVE
                asyncio.run_coroutine_threadsafe(wamp_checks(session), loop)

                # LOADING BOARD MODULES
                try:

                    modulesLoader(session)

                except Exception as e:
                    LOG.warning("WARNING - Could not load modules: " + str(e))
                    utils.LR_restart()

                # Reset flag to False
                # reconnection = False

            else:
                Bye()

    except exception.ApplicationError as e:
        LOG.error(" - Iotronic Connection RPC error: " + str(e))
        # Iotronic is offline the board can not call
        # the "stack4things.connection" RPC.
        # The board will disconnect from WAMP agent and retry later.
        reconnection = True

        # We restart Lightning-rod if RPC 'stack4things.connection' is not
        # available, this means Wagent is unreachable
        utils.LR_restart()

    except Exception as e:
        LOG.warning("Iotronic board connection error: " + str(e))


def wampConnect(wamp_conf):
    """WAMP connection procedures.

    :param wamp_conf: WAMP configuration from settings.json file

    """

    LOG.info("WAMP connection precedures:")

    try:

        LOG.info(
            "WAMP status @ boot:" +
            "\n- board = " + str(board.status) +
            "\n- reconnection = " + str(reconnection) +
            "\n- connected = " + str(connected)
        )

        wamp_transport = wamp_conf['url']
        wurl_list = wamp_transport.split(':')
        is_wss = False

        if wurl_list[0] == "wss":
            is_wss = True
        whost = wurl_list[1].replace('/', '')
        wport = int(wurl_list[2].replace('/', ''))

        if is_wss and CONF.skip_cert_verify:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            wamp_transport = [
                {
                    "url": wamp_transport,
                    "serializers": ["json"],
                    "endpoint": {
                        "type": "tcp",
                        "host": whost,
                        "port": wport,
                        "tls": ctx
                    },
                },
            ]

        # LR creates the Autobahn Asyncio Component that points to the
        # WAMP Agent (main/registration agent)
        global component
        component = Component(
            transports=wamp_transport,
            realm=wamp_conf['realm']
        )

        # To manage the registration stage: we got the info for the main
        # WAMP agent and LR is going to connect to it starting the Component
        # with the new WAMP configuration.
        if connected == False and board.status == "registered" \
                and reconnection == False:
            component.start(loop)

        @component.on_join
        async def join(session, details):
            """Execute the following procedures when the board connects
            to Crossbar.

            :param details: WAMP session details

            """

            print("WAMP SOCKET: " + str(psutil.Process().connections()[0]))

            global connected
            connected = True

            # LIGHTNING-ROD STATES:
            # - REGISTRATION STATE: the first connection to Iotronic
            # - FIRST CONNECTION: the board become operative after registration
            # - LIGHTNING-ROD BOOT: the first connection to WAMP
            #                       after Lightning-rod starting
            # - WAMP RECOVERY: when the established WAMP connection fails

            global reconnection

            # reconnection flag is False when the board is:
            # - LIGHTNING-ROD BOOT
            # - REGISTRATION STATE
            # - FIRST CONNECTION
            #
            # reconnection flag is True when the board is:
            # - WAMP RECOVERY

            global SESSION
            SESSION = session

            # LOG.debug(" - session: " + str(details))

            board.session_id = details.session

            LOG.info(" - Joined in realm " + board.wamp_config['realm'] + ":")
            LOG.info("   - WAMP Agent: " + str(board.agent))
            print(" - WAMP Agent: " + str(board.agent) + " - "
                  + str(wamp_conf['url']))
            LOG.info("   - Session ID: " + str(board.session_id))
            print(" - Session ID: " + str(board.session_id))
            LOG.info("   - Board status:  " + str(board.status))

            if reconnection is False:

                if board.uuid is None:

                    ######################
                    # REGISTRATION STATE #
                    ######################
                    # If in the LR configuration file there is not the
                    # Board UUID specified it means the board is a new one
                    # and it has to call IoTronic in order to complete
                    # the registration.

                    try:

                        LOG.info(" - Board needs to be registered.")

                        rpc = u'stack4things.register'

                        with timeoutRPC(seconds=5, action=rpc):
                            res = await session.call(
                                rpc,
                                code=board.code,
                                session=board.session_id
                            )

                        w_msg = WM.deserialize(res)

                        # LOG.info(" - Board registration result: \n" +
                        #         json.loads(w_msg.message, indent=4))

                        if w_msg.result == WM.SUCCESS:

                            LOG.info("Registration authorized by IoTronic:\n"
                                     + str(w_msg.message))

                            # the 'message' field contains
                            # the board configuration to load
                            board.setConf(w_msg.message)

                            # We need to disconnect the client from the
                            # registration-agent in order to reconnect
                            # to the WAMP agent assigned by Iotronic
                            # at the provisioning stage
                            LOG.info(
                                "\n\nDisconnecting from Registration Agent "
                                "to load new settings...\n\n")

                            # We restart Lightning-rod if RPC
                            # 'stack4things.connection' is not available,
                            # this means Wagent is unreachable
                            utils.LR_restart()

                        else:
                            LOG.error("Registration denied by Iotronic: "
                                      + str(w_msg.message))
                            Bye()

                    except exception.ApplicationError as e:
                        LOG.error("IoTronic registration error: " + str(e))
                        # Iotronic is offline the board can not call the
                        # "stack4things.connection" RPC. The board will
                        # disconnect from WAMP agent and retry later.

                        # TO ACTIVE BOOT CONNECTION RECOVERY MODE
                        reconnection = True

                        # We restart Lightning-rod if RPC
                        # 'stack4things.connection' is not available,
                        # this means Wagent is unreachable
                        utils.LR_restart()

                    except Exception as e:
                        LOG.warning(
                            " - Board registration call error: " + str(e))
                        Bye()

                else:

                    if board.status == "registered":
                        ####################
                        # FIRST CONNECTION #
                        ####################

                        # In this case we manage the first connection
                        # after the registration stage:
                        # Lightining-rod sets its status to "operative"
                        # completing the provisioning and configuration stage.
                        LOG.info("\n\n\nBoard is becoming operative...\n\n\n")
                        board.updateStatus("operative")
                        board.loadSettings()
                        LOG.info("WAMP status @ first connection:" +
                                 "\n- board = " + str(board.status) +
                                 "\n- reconnection = " + str(reconnection) +
                                 "\n- connected = " + str(connected)
                                 )
                        await IotronicLogin(board, session, details)

                    elif board.status == "operative":
                        ######################
                        # LIGHTNING-ROD BOOT #
                        ######################

                        # After join to WAMP agent, Lightning-rod will:
                        # - authenticate to Iotronic
                        # - load the enabled modules

                        # The board will keep at this stage until
                        # it will succeed to connect to Iotronic.
                        await IotronicLogin(board, session, details)

                    else:
                        LOG.error("Wrong board status '" + board.status + "'.")
                        Bye()

            else:

                #################
                # WAMP RECOVERY #
                #################

                LOG.info("IoTronic connection recovery:")

                try:

                    rpc = str(board.agent) + u'.stack4things.connection'

                    with timeoutRPC(seconds=CONF.connection_timer, action=rpc):
                        res = await session.call(
                            rpc,
                            uuid=board.uuid,
                            session=details.session
                        )

                    w_msg = WM.deserialize(res)

                    if w_msg.result == WM.SUCCESS:

                        LOG.info(" - Access granted to Iotronic (recovery).")

                        # LOADING BOARD MODULES
                        # If the board is in WAMP connection recovery state
                        # we need to register again the RPCs of each module
                        try:

                            moduleReloadInfo(session)

                            # Reset flag to False
                            reconnection = False

                            LOG.info("WAMP Session Recovered!")

                            LOG.info("\n\nListening...\n\n")

                            # WS ALIVE
                            asyncio.run_coroutine_threadsafe(
                                wamp_checks(session),
                                loop
                            )

                        except Exception as e:
                            LOG.warning(
                                "WARNING - Could not reload modules: "
                                + str(e))
                            Bye()

                    else:
                        LOG.error("Access to IoTronic denied: "
                                  + str(w_msg.message))
                        Bye()

                except exception.ApplicationError as e:
                    LOG.error("IoTronic connection error:\n" + str(e))
                    # Iotronic is offline the board can not call
                    # the "stack4things.connection" RPC.
                    # The board will disconnect from WAMP agent and retry later

                    # TO ACTIVE WAMP CONNECTION RECOVERY MODE
                    reconnection = False

                    # We restart Lightning-rod if RPC 'stack4things.connection'
                    # is not available, this means Wagent is unreachable
                    utils.LR_restart()

                except Exception as e:
                    LOG.warning("Board connection error after WAMP recovery: "
                                + str(e))
                    Bye()

        @component.on_leave
        async def onLeave(session, details):
            LOG.warning("WAMP Session Left: reason = " + str(details.reason))

        @component.on_disconnect
        async def onDisconnect(session, was_clean):
            """Procedure triggered on WAMP connection lost.
            :param session:
            :param was_clean:
            :return:
            """

            LOG.warning('WAMP Transport Left: was_clean = ' + str(was_clean))
            global connected
            connected = False

            global reconnection

            LOG.info(
                "WAMP status on disconnect:" +
                "\n- board = " + str(board.status) +
                "\n- reconnection = " + str(reconnection) +
                "\n- connected = " + str(connected)
            )

            if board.status == "operative" and reconnection is False:

                #################
                # WAMP RECOVERY #
                #################
                # we need to recover wamp session and
                # we set reconnection flag to True in order to activate
                # the module-RPCs registration procedure for each module

                reconnection = True

                # LR needs to reconncet to WAMP
                if not connected:
                    LOG.warning(".............WAMP DISCONNECTION.............")
                    LOG.info(
                        "WAMP status on disconnect:" +
                        "\n- board = " + str(board.status) +
                        "\n- reconnection = " + str(reconnection) +
                        "\n- connected = " + str(connected)
                    )

                    # component.start(loop)

            elif board.status == "operative" and reconnection is True:

                ######################
                # LIGHTNING-ROD BOOT #
                ######################
                # At this stage if the reconnection flag was set to True
                # it means that we forced the reconnection procedure
                # because of the board is not able to connect to IoTronic
                # calling "stack4things.connection" RPC...
                # it means IoTronic is offline!

                # We need to reset the reconnection flag to False in order to
                # do not enter in module-RPCs registration procedure...
                # At this stage the board tries to reconnect to
                # IoTronic until it will come online again.
                reconnection = False

                # LR needs to reconncet to WAMP
                LOG.warning(".............WAMP DISCONNECTION.............")
                LOG.info("WAMP status on disconnect:" +
                         "\n- board = " + str(board.status) +
                         "\n- reconnection = " + str(reconnection) +
                         "\n- connected = " + str(connected)
                         )

                # component.start(loop)

            elif (board.status == "registered"):
                ######################
                # REGISTRATION STATE #
                ######################

                # LR was disconnected from Registration Agent
                # in order to connect it to the assigned WAMP Agent.

                LOG.debug("\n\nReconnecting after registration...\n\n")

                # LR load the new configuration and gets the new WAMP Agent
                board.loadSettings()

                # LR has to connect to the assigned WAMP Agent
                wampConnect(board.wamp_config)

            else:
                LOG.error("Reconnection wrong status!")

    except Exception as err:
        LOG.error(" - WAMP connection error: " + str(err))
        Bye()


def moduleWampRegister(session, meth_list):
    """This function register for each module methods the relative RPC.

    :param session:
    :param meth_list:

    """

    if len(meth_list) == 2:

        LOG.info("   - No procedures to register!")

    else:

        for meth in meth_list:
            # We don't considere the "__init__", "finalize" and
            # "restore" methods
            if (meth[0] != "__init__") & (meth[0] != "finalize") \
                    & (meth[0] != "restore"):

                rpc_addr = u'iotronic.' + str(board.session_id) + '.' + \
                           board.uuid + '.' + meth[0]

                if not meth[0].startswith('_'):
                    session.register(meth[1], rpc_addr)
                    LOG.info("   --> " + str(meth[0]))


def modulesLoader(session):
    """Modules loader method thorugh stevedore libraries.

    :param session:

    """

    LOG.info("Available modules: ")

    ep = []

    for ep in pkg_resources.iter_entry_points(group='s4t.modules'):
        LOG.info(" - " + str(ep))

    if not ep:

        LOG.info("No modules available!")
        sys.exit()

    else:

        modules = extension.ExtensionManager(
            namespace='s4t.modules',
            # invoke_on_load=True,
            # invoke_args=(session,),
        )

        LOG.info('Modules to load:')

        for ext in modules.extensions:

            # LOG.debug(ext.name)

            if (ext.name == 'gpio') & (board.type == 'server'):
                LOG.info("- GPIO module disabled for 'server' devices")

            else:
                mod = ext.plugin(board, session)

                global MODULES
                MODULES[mod.name] = mod

                # Methods list for each module
                meth_list = inspect.getmembers(mod, predicate=inspect.ismethod)

                global RPC
                RPC[mod.name] = meth_list

                if len(meth_list) == 3:
                    # there are at least two methods for each module:
                    # "__init__" and "finalize"

                    LOG.info(" - No RPC to register for "
                             + str(ext.name) + " module!")

                else:
                    LOG.info(" - RPC list of " + str(mod.name) + ":")
                    moduleWampRegister(SESSION, meth_list)

                # Call the finalize procedure for each module
                mod.finalize()

        LOG.info("Lightning-rod modules loaded.")
        LOG.info("\n\nListening...")


def moduleReloadInfo(session):
    """This function is used in the reconnection stage to register

    again the RPCs of each module and for device.

    :param session: WAMP session object.

    """

    LOG.info("\n\nModules reloading after WAMP recovery...\n\n")

    try:

        # Call module restore procedures and
        # register RPCs for each Lightning-rod module
        for mod_name in MODULES:
            LOG.info("- Registering RPCs for module " + str(mod_name))
            moduleWampRegister(session, RPC[mod_name])

            LOG.info("- Restoring module " + str(mod_name))
            MODULES[mod_name].restore()

        # Register RPCs for the device
        for dev in RPC_devices:
            LOG.info("- Registering RPCs for device " + str(dev))
            moduleWampRegister(session, RPC_devices[dev])

    except Exception as err:
        LOG.warning("Board modules reloading error: " + str(err))
        utils.LR_restart()


def Bye():
    LOG.info("Bye!")
    os._exit(1)


def LogoLR():
    LOG.info('##############################')
    LOG.info('  Stack4Things Lightning-rod')
    LOG.info('##############################')


def main():
    LightningRod()



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import errno
import json
import os
import psutil
import pyinotify
import signal
import subprocess
import time

from datetime import datetime
from threading import Thread
from urllib.parse import urlparse

from iotronic_lightningrod.config import package_path
from iotronic_lightningrod.modules import Module
from iotronic_lightningrod.modules import utils

import iotronic_lightningrod.wampmessage as WM

from iotronic_lightningrod import lightningrod


from oslo_config import cfg
from oslo_log import log as logging
LOG = logging.getLogger(__name__)


wstun_opts = [
    cfg.StrOpt(
        'wstun_bin',
        default='/usr/bin/wstun',
        help=('WSTUN bin for Services Manager')
    ),
]

CONF = cfg.CONF

service_group = cfg.OptGroup(
    name='services', title='Services options'
)
CONF.register_group(service_group)
CONF.register_opts(wstun_opts, group=service_group)


s_conf_FILE = CONF.lightningrod_home + "/services.json"


class ServiceManager(Module.Module):

    def __init__(self, board, session):
        super(ServiceManager, self).__init__("ServiceManager", board)

        print("\nWSTUN bin path: " + str(CONF.services.wstun_bin))

        self.wstun_ip = urlparse(board.wamp_config["url"])[1].split(':')[0]
        self.wstun_port = "8080"

        is_wss = False
        wurl_list = board.wamp_config["url"].split(':')
        if wurl_list[0] == "wss":
            is_wss = True

        if is_wss:
            self.wstun_url = "wss://" + self.wstun_ip + ":" + self.wstun_port
        else:
            self.wstun_url = "ws://" + self.wstun_ip + ":" + self.wstun_port

    def finalize(self):
        LOG.info("Cloud service tunnels to initialization:")

        # Load services.json configuration file
        s_conf = self._loadServicesConf()

        if s_conf == None:

            LOG.error(" --> Error loading services.json file: " +
                      "backup is not restorable!")

            path_services_template = \
                package_path + "/templates/services.example.json"

            if os.path.isfile(path_services_template):

                LOG.info(" --> restoring services.json template")

                # Restore services.json template file
                os.system(
                    'cp ' + path_services_template + ' ' + s_conf_FILE
                )

                LOG.warn(" --> template restored: services configuration " +
                         "must be injected from Iotronic.")
            else:
                LOG.error(" --> services.json template does not exist!")

        else:

            if len(s_conf['services']) != 0:

                wstun_process_list = []

                for p in psutil.process_iter():
                    if len(p.cmdline()) != 0:
                        if (p.name() == "node" and "wstun" in p.cmdline()[1]):
                            wstun_process_list.append(p)

                if len(s_conf) != 0:
                    print("\nWSTUN processes:")

                for s_uuid in s_conf['services']:

                    service_name = \
                        s_conf['services'][s_uuid]['name']
                    service_pid = \
                        s_conf['services'][s_uuid]['pid']
                    LOG.info(" - " + service_name)

                    if len(wstun_process_list) != 0:

                        for wp in wstun_process_list:

                            if service_pid == wp.pid:
                                LOG.info(
                                    " --> the tunnel for '" + service_name
                                    + "' already exists; killing..."
                                )

                                # 1. Kill wstun process (if exists)

                                # No zombie alert activation
                                lightningrod.zombie_alert = False
                                LOG.debug(
                                    "[WSTUN-RESTORE] - "
                                    "on-finalize zombie_alert: " +
                                    str(lightningrod.zombie_alert)
                                )

                                try:
                                    os.kill(service_pid, signal.SIGINT)
                                    print("OLD WSTUN KILLED: " + str(wp))
                                    LOG.info(" --> service '" + service_name
                                             + "' with PID " + str(service_pid)
                                             + " was killed; "
                                             + "creating new one...")

                                except OSError:
                                    LOG.warning(
                                        " - WSTUN process already killed, "
                                        "creating new one...")

                                break

                    # 2. Create the reverse tunnel
                    public_port = \
                        s_conf['services'][s_uuid]['public_port']
                    local_port = \
                        s_conf['services'][s_uuid]['local_port']

                    wstun = self._startWstun(
                        public_port, local_port, event="boot"
                    )

                    if wstun != None:

                        service_pid = wstun.pid

                        # 3. Update services.json file
                        s_conf['services'][s_uuid]['pid'] = \
                            service_pid
                        s_conf['services'][s_uuid]['updated_at'] = \
                            datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')

                        self._updateServiceConf(s_conf, s_uuid,
                                                output=True)

                        LOG.info(" --> Cloud service '" + service_name
                                 + "' tunnel established.")
                    else:
                        message = "Error spawning " + str(service_name) \
                                  + " service tunnel!"
                        LOG.error(" - " + message)

                    signal.signal(signal.SIGCHLD, self._zombie_hunter)

                # Reactivate zombies monitoring
                if not lightningrod.zombie_alert:
                    lightningrod.zombie_alert = True

            else:
                LOG.info(" --> No service tunnels to establish.")

                signal.signal(signal.SIGCHLD, self._zombie_hunter)

    def restore(self):
        LOG.info("Cloud service tunnels to restore:")

        # Load services.json configuration file
        s_conf = self._loadServicesConf()

        if s_conf == None:

            LOG.error(" --> Error loading services.json file: "
                      "backup is not restorable!")

        else:

            if len(s_conf['services']) != 0:

                wstun_process_list = []

                # No zombie alert activation
                lightningrod.zombie_alert = False
                LOG.debug("[WSTUN-RESTORE] - Restore zombie_alert: " + str(
                    lightningrod.zombie_alert))

                # Collect all alive WSTUN proccesses
                for p in psutil.process_iter():
                    if (p.name() == "node"):
                        if (p.status() == psutil.STATUS_ZOMBIE):
                            LOG.warning("WSTUN ZOMBIE: " + str(p))
                            wstun_process_list.append(p)
                        elif ("wstun" in p.cmdline()[1]):
                            LOG.warning("WSTUN ALIVE: " + str(p))
                            wstun_process_list.append(p)

                            psutil.Process(p.pid).kill()
                            LOG.warning(" --> PID " + str(p.pid) + " killed!")

                LOG.debug("[WSTUN-RESTORE] - WSTUN processes to restore:\n"
                          + str(wstun_process_list))

                for s_uuid in s_conf['services']:

                    Thread(
                        target=self._restoreWSTUN,
                        args=(s_conf, s_uuid,)
                    ).start()
                    time.sleep(2)

                # Reactivate zombies monitoring
                if not lightningrod.zombie_alert:
                    lightningrod.zombie_alert = True

            else:
                LOG.info(" --> No service tunnels to restore.")

    def _zombie_hunter(self, signum, frame):

        wstun_found = False

        if (lightningrod.zombie_alert):
            # print(signum); traceback.print_stack(frame)

            zombie_list = []

            for p in psutil.process_iter():
                if len(p.cmdline()) == 0:
                    if ((p.name() == "node") and
                            (p.status() == psutil.STATUS_ZOMBIE)):
                        print(" - process: " + str(p))
                        zombie_list.append(p.pid)

            if len(zombie_list) == 0:
                # print(" - no action required.")
                return

            print("\nCheck killed process...")
            print(" - Zombies found: " + str(zombie_list))

            # Load services.json configuration file
            s_conf = self._loadServicesConf()

            if s_conf == None:
                LOG.error(" --> Error loading services.json file: "
                          "backup is not restorable!")

            else:

                for s_uuid in s_conf['services']:

                    service_pid = s_conf['services'][s_uuid]['pid']

                    if service_pid in zombie_list:

                        message = "WSTUN zombie process ALERT!"
                        print(" - " + str(message))
                        LOG.debug("[WSTUN-RESTORE] --> " + str(message))

                        wstun_found = True

                        print(s_conf['services'][s_uuid])
                        service_public_port = \
                            s_conf['services'][s_uuid]['public_port']
                        service_local_port = \
                            s_conf['services'][s_uuid]['local_port']
                        service_name = \
                            s_conf['services'][s_uuid]['name']

                        try:

                            wstun = self._startWstun(
                                service_public_port,
                                service_local_port,
                                event="zombie"
                            )

                            if wstun != None:
                                service_pid = wstun.pid

                                # UPDATE services.json file
                                s_conf['services'][s_uuid]['pid'] = \
                                    service_pid
                                s_conf['services'][s_uuid]['updated_at'] = \
                                    datetime.now().strftime(
                                        '%Y-%m-%dT%H:%M:%S.%f'
                                    )

                                self._updateServiceConf(
                                    s_conf,
                                    s_uuid,
                                    output=True
                                )

                                message = "Zombie service " \
                                          + str(service_name) \
                                          + " restored on port " \
                                          + str(service_public_port) \
                                          + " on " + self.wstun_ip
                                LOG.info(" - " + message
                                         + " with PID " + str(service_pid))

                        except Exception:
                            pass

                        break

                if not wstun_found:
                    message = "Tunnel killed by LR"
                    print(" - " + str(message))
                    # LOG.debug("[WSTUN-RESTORE] --> " + str(message))

        else:
            print("WSTUN kill event:")
            message = "Tunnel killed by LR"
            print(" - " + str(message))
            # LOG.debug("[WSTUN-RESTORE] --> " + str(message))
            # lightningrod.zombie_alert = True

    def _restoreWSTUN(self, s_conf, s_uuid):
        service_name = s_conf['services'][s_uuid]['name']
        service_pid = s_conf['services'][s_uuid]['pid']
        LOG.info(" - " + service_name)

        # Create the reverse tunnel again
        public_port = \
            s_conf['services'][s_uuid]['public_port']
        local_port = \
            s_conf['services'][s_uuid]['local_port']

        wstun = self._startWstun(public_port, local_port, event="restore")

        if wstun != None:
            service_pid = wstun.pid

            # 3. Update services.json file
            s_conf['services'][s_uuid]['pid'] = \
                service_pid
            s_conf['services'][s_uuid]['updated_at'] = \
                datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')

            self._updateServiceConf(
                s_conf,
                s_uuid,
                output=True
            )

            LOG.info(" --> Cloud service '" + service_name
                     + "' tunnel restored.")
        else:
            message = "Error spawning " + str(service_name) \
                      + " service tunnel!"
            LOG.error(" - " + message)

    def _loadServicesConf(self):
        """Load services.json JSON configuration.

        :return: JSON Services configuration

        """

        try:

            with open(s_conf_FILE) as settings:
                s_conf = json.load(settings)

        except Exception as err:
            LOG.error(" --> Parsing error in "
                      + s_conf_FILE + ": " + str(err))

            if os.path.isfile(s_conf_FILE):

                LOG.info(" --> restoring services.json file...")

                # Restore backup json file on error
                os.system(
                    'cp ' + s_conf_FILE + '.bkp ' + s_conf_FILE
                )

                try:
                    with open(s_conf_FILE) as settings:
                        s_conf = json.load(settings)
                except Exception as err:
                    LOG.error(" --> Parsing backup file error in " +
                              s_conf_FILE + ": " + str(err))
                    s_conf = None

            else:
                LOG.error(" --> services.json backup file does not exist!")
                s_conf = None

        return s_conf

    def _wstunMon(self, wstun):

        wfd_check = True

        while (wfd_check):
            try:

                wp = psutil.Process(int(wstun.pid))
                wstun_fd = wp.connections()[0].fd
                if len(wp.connections()) != 0:
                    LOG.debug("WSTUN alive socket: " + str(wp.connections()))
                wfd_check = False

            except IndexError as err:
                # LOG.error(str(err) + " - RETRY...")
                pass

            time.sleep(1)

        class EventProcessor(pyinotify.ProcessEvent):
            _methods = [
                # "IN_CREATE",
                # "IN_OPEN",
                # "IN_ACCESS",
                # "IN_ATTRIB",
                "IN_CLOSE_NOWRITE",
                "IN_CLOSE_WRITE",
                "IN_DELETE",
                "IN_DELETE_SELF",
                # "IN_IGNORED",
                # "IN_MODIFY",
                # "IN_MOVE_SELF",
                # "IN_MOVED_FROM",
                # "IN_MOVED_TO",
                # "IN_Q_OVERFLOW",
                # "IN_UNMOUNT",
                "default"
            ]

        def process_generator(cls, method):
            def _method_name(self, event):
                if(event.maskname == "IN_CLOSE_WRITE"):
                    LOG.info("WSTUN fd socket closed: " + str(event.pathname))
                    LOG.debug(
                        " - FD change notify:"
                        + "\nmethod: process_{}()\n"
                        + "file_path: {}\n"
                        + "event: {}\n".format(
                            method,
                            event.pathname,
                            event.maskname
                        )
                    )

                    os.kill(wstun.pid, signal.SIGKILL)

            _method_name.__name__ = "process_{}".format(method)
            setattr(cls, _method_name.__name__, _method_name)

        for method in EventProcessor._methods:
            process_generator(EventProcessor, method)

        watch_manager = pyinotify.WatchManager()
        event_notifier = pyinotify.ThreadedNotifier(
            watch_manager, EventProcessor()
        )

        watch_this = os.path.abspath(
            "/proc/" + str(wstun.pid) + "/fd/" + str(wstun_fd)
        )
        watch_manager.add_watch(watch_this, pyinotify.ALL_EVENTS)
        event_notifier.start()

    def _startWstun(self, public_port, local_port, event="no-set"):

        import socket

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(4)
        result = sock.connect_ex((self.wstun_ip, int(self.wstun_port)))

        if result == 0:

            sock.close()  # close check socket

            opt_reverse = "-r" + str(public_port) + ":127.0.0.1:" + str(
                local_port)

            try:
                wstun = subprocess.Popen(
                    [CONF.services.wstun_bin, opt_reverse, self.wstun_url],
                    stdout=subprocess.PIPE
                )

                if (event != "boot"):
                    print("WSTUN start event:")

                cmd_print = 'WSTUN exec: ' + str(CONF.services.wstun_bin) \
                            + opt_reverse + ' ' + self.wstun_url
                print(" - " + str(cmd_print))
                LOG.debug(cmd_print)

                # WSTUN MON
                # #############################################################

                Thread(
                    target=self._wstunMon,
                    args=(wstun,)
                ).start()

                # self._wstunMon(wstun)

                # #############################################################

            except Exception as err:
                LOG.error("Error spawning WSTUN process: " + str(err))
                wstun = None

        else:
            LOG.error("Error spawning WSTUN process: WSTUN server is offline!")
            wstun = None

        return wstun

    def _updateServiceConf(self, s_conf, s_uuid, output=True):

        # Apply the changes to services.json
        with open(s_conf_FILE, 'w') as f:
            json.dump(s_conf, f, indent=4)

            if output:
                LOG.info(" - service updated:\n" + json.dumps(
                    s_conf['services'][s_uuid],
                    indent=4,
                    sort_keys=True
                ))
            else:
                LOG.info(" - services.json file updated!")

        # Backup json file before update
        os.system(
            'cp ' + s_conf_FILE + ' ' + s_conf_FILE + '.bkp'
        )

    async def ServiceEnable(self, service, public_port):

        rpc_name = utils.getFuncName()

        service_name = service['name']
        s_uuid = service['uuid']
        local_port = service['port']

        LOG.info("RPC " + rpc_name + " CALLED for '" + service_name
                 + "' (" + s_uuid + ") service:")

        try:

            wstun = self._startWstun(public_port, local_port, event="enable")

            if wstun != None:

                service_pid = wstun.pid

                # Load services.json configuration file
                s_conf = self._loadServicesConf()

                if s_conf == None:
                    message = "Error loading services.json file: " \
                              + "backup is not restorable!"

                    LOG.error(" --> " + message)
                    w_msg = WM.WampError(message)

                else:

                    # Save plugin settings in services.json
                    if s_uuid not in s_conf['services']:

                        # It is a new plugin
                        s_conf['services'][s_uuid] = {}
                        s_conf['services'][s_uuid]['name'] = \
                            service_name
                        s_conf['services'][s_uuid]['public_port'] = \
                            public_port
                        s_conf['services'][s_uuid]['local_port'] = \
                            local_port
                        s_conf['services'][s_uuid]['pid'] = \
                            service_pid
                        s_conf['services'][s_uuid]['enabled_at'] = \
                            datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
                        s_conf['services'][s_uuid]['updated_at'] = ""

                    else:
                        # The service was already added and we are updating it
                        s_conf['services'][s_uuid]['updated_at'] = \
                            datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
                        LOG.info(" - services.json file updated!")

                    # Apply the changes to services.json
                    self._updateServiceConf(s_conf, s_uuid, output=True)

                    message = "Cloud service '" + str(service_name) \
                              + "' exposed on port " \
                              + str(public_port) + " on " + self.wstun_ip

                    LOG.info(" - " + message + " with PID " + str(service_pid))

                    w_msg = WM.WampSuccess(message)

            else:
                message = "Error spawning " + str(service_name) \
                          + " service tunnel!"
                LOG.error(" - " + message)
                w_msg = WM.WampError(message)

        except Exception as err:
            message = "Error exposing " + str(service_name) \
                      + " service: " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(message)

        return w_msg.serialize()

    async def ServiceDisable(self, service):

        rpc_name = utils.getFuncName()

        service_name = service['name']
        s_uuid = service['uuid']

        LOG.info("RPC " + rpc_name
                 + " CALLED for '" + service_name
                 + "' (" + s_uuid + ") service:")

        # Remove from services.json file
        try:

            # Load services.json configuration file
            s_conf = self._loadServicesConf()

            if s_conf == None:
                LOG.error(" --> Error loading services.json file: "
                          "backup is not restorable!")

                message = "Error loading services.json file: " \
                          + "backup is not restorable!"
                LOG.error(" --> " + message)
                w_msg = WM.WampError(message)

            else:

                if s_uuid in s_conf['services']:

                    service_pid = \
                        s_conf['services'][s_uuid]['pid']

                    try:

                        # No zombie alert activation
                        lightningrod.zombie_alert = False

                        os.kill(service_pid, signal.SIGKILL)

                        message = "Cloud service '" \
                                  + str(service_name) + "' tunnel disabled."

                        del s_conf['services'][s_uuid]

                        self._updateServiceConf(s_conf, s_uuid,
                                                output=False)

                        LOG.info(" - " + message)

                        # Reactivate zombies monitoring
                        if not lightningrod.zombie_alert:
                            lightningrod.zombie_alert = True

                        w_msg = WM.WampSuccess(message)

                    except Exception as err:
                        if err.errno == errno.ESRCH:  # No such process
                            message = "Service '" + str(service_name) \
                                      + "' WSTUN process is not running!"
                            LOG.warning(" - " + message)

                            del s_conf['services'][s_uuid]

                            self._updateServiceConf(
                                s_conf,
                                s_uuid,
                                output=False
                            )

                            # Reactivate zombies monitoring
                            if not lightningrod.zombie_alert:
                                lightningrod.zombie_alert = True

                            w_msg = WM.WampWarning(message)

                        else:

                            message = "Error disabling '" + str(
                                service_name) + "' service tunnel: " + str(err)
                            LOG.error(" - " + message)

                            # Reactivate zombies monitoring
                            if not lightningrod.zombie_alert:
                                lightningrod.zombie_alert = True

                            w_msg = WM.WampError(message)

                else:
                    message = rpc_name + " result:  " + s_uuid \
                        + " already removed!"
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

        except Exception as err:
            message = "Updating services.json error: " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(message)

        return w_msg.serialize()

    async def ServiceRestore(self, service, public_port):

        rpc_name = utils.getFuncName()

        service_name = service['name']
        s_uuid = service['uuid']

        LOG.info("RPC " + rpc_name
                 + " CALLED for '" + service_name
                 + "' (" + s_uuid + ") service:")

        # Load services.json configuration file
        s_conf = self._loadServicesConf()

        if s_conf == None:

            message = "Error loading services.json file: " \
                      + "backup is not restorable!"
            LOG.error(" --> " + message)
            w_msg = WM.WampError(message)

        else:

            if s_uuid in s_conf['services']:

                local_port = \
                    s_conf['services'][s_uuid]['local_port']
                service_pid = \
                    s_conf['services'][s_uuid]['pid']

                try:

                    # No zombie alert activation
                    lightningrod.zombie_alert = False

                    # 1. Kill wstun process (if exists)
                    try:
                        os.kill(service_pid, signal.SIGKILL)
                        LOG.info(" - service '" + service_name
                                 + "' with PID " + str(service_pid)
                                 + " was killed.")
                    except OSError:
                        LOG.warning(" - WSTUN process already killed: "
                                    "creating new one...")

                    # 2. Create the reverse tunnel
                    wstun = self._startWstun(
                        public_port,
                        local_port,
                        event="kill-restore"
                    )

                    if wstun != None:
                        service_pid = wstun.pid

                        # UPDATE services.json file
                        s_conf['services'][s_uuid]['pid'] = \
                            service_pid
                        s_conf['services'][s_uuid]['updated_at'] = \
                            datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')

                        self._updateServiceConf(s_conf, s_uuid,
                                                output=True)

                        message = "service " + str(service_name) \
                                  + " restored on port " \
                                  + str(public_port) + " on " + self.wstun_ip
                        LOG.info(" - " + message
                                 + " with PID " + str(service_pid))

                        # Reactivate zombies monitoring
                        if not lightningrod.zombie_alert:
                            lightningrod.zombie_alert = True

                        w_msg = WM.WampSuccess(message)

                    else:
                        message = "Error spawning " + str(service_name) \
                                  + " service tunnel!"
                        LOG.error(" - " + message)

                        # Reactivate zombies monitoring
                        if not lightningrod.zombie_alert:
                            lightningrod.zombie_alert = True

                        w_msg = WM.WampError(message)

                except Exception as err:
                    message = "Error restoring '" + str(service_name) \
                              + "' service tunnel: " + str(err)
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

            else:

                local_port = service['port']

                wstun = self._startWstun(
                    public_port,
                    local_port,
                    event="clean-restore"
                )

                if wstun != None:

                    service_pid = wstun.pid

                    s_conf['services'][s_uuid] = {}
                    s_conf['services'][s_uuid]['name'] = \
                        service_name
                    s_conf['services'][s_uuid]['public_port'] = \
                        public_port
                    s_conf['services'][s_uuid]['local_port'] = \
                        local_port
                    s_conf['services'][s_uuid]['pid'] = \
                        service_pid
                    s_conf['services'][s_uuid]['enabled_at'] = \
                        datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
                    s_conf['services'][s_uuid]['updated_at'] = ""

                    self._updateServiceConf(s_conf, s_uuid,
                                            output=True)

                    message = "service " + str(service_name) \
                              + " restored on port " \
                              + str(public_port) + " on " + self.wstun_ip
                    LOG.info(" - " + message + " with PID " + str(service_pid))

                    w_msg = WM.WampSuccess(message)

                else:
                    message = "Error spawning " + str(service_name) \
                              + " service tunnel!"
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

        return w_msg.serialize()



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

from iotronic_lightningrod.devices import Device
from iotronic_lightningrod.devices.gpio import yun

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

# Linino references: http://wiki.linino.org/doku.php?id=wiki:lininoio_sysfs


class System(Device.Device):

    def __init__(self):
        super(System, self).__init__("yun")

        self.gpio = yun.YunGpio()

        self.gpio.EnableGPIO()

    def finalize(self):
        """Function called at the end of module loading (after RPC registration).

        :return:

        """
        pass

    async def testLED(self):
        LOG.info(" - testLED CALLED...")

        await self.gpio.blinkLed()

        result = "testLED: LED blinking!\n"
        LOG.info(result)
        return result

    async def setGPIOs(self, Dpin, direction, value):

        LOG.info(" - setGPIOs CALLED... digital pin " + Dpin
                 + " (GPIO n. " + self.gpio.MAPPING[Dpin] + ")")

        result = await self.gpio._setGPIOs(Dpin, direction, value)
        LOG.info(result)
        return result

    async def readVoltage(self, Apin):
        """To read the voltage applied on the pin A0,A1,A2,A3,A4,A5

        """
        LOG.info(" - readVoltage CALLED... reading pin " + Apin)

        voltage = self.gpio._readVoltage(Apin)

        result = await "read voltage for " + Apin + " pin: " + voltage
        LOG.info(result)
        return result



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import abc
import six
import threading

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Plugin(threading.Thread):

    def __init__(self, uuid, name, q_result=None, params=None):

        threading.Thread.__init__(self)
        # self.setDaemon(1)
        self.setName("Plugin " + str(self.name))  # Set thread name
        LOG.debug("Plugin Name: " + self.name)

        self.uuid = uuid
        self.name = name
        self.status = "INITED"
        self.setStatus(self.status)
        self._is_running = True
        self.params = params
        self.q_result = q_result
        self.type = type

    @abc.abstractmethod
    def run(self):
        """RUN method where to implement the user's plugin logic

        """
    def stop(self):
        self._is_running = False

    def checkStatus(self):
        # LOG.debug("Plugin " + self.name + " check status: " + self.status)
        return self.status

    def setStatus(self, status):
        self.status = status
        # LOG.debug("Plugin " + self.name + " changed status: " + self.status)

    def complete(self, rpc_name, result):
        self.setStatus(result)
        result = rpc_name + " result: " + self.checkStatus()

        return result



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import inspect

from iotronic_lightningrod.devices import Device
from iotronic_lightningrod.devices.gpio import server

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


def whoami():
    return inspect.stack()[1][3]


def makeNothing():
    pass


class System(Device.Device):

    def __init__(self):
        super(System, self).__init__("server")

        server.ServerGpio().EnableGPIO()

    def finalize(self):
        """Function called at the end of module loading (after RPC registration).

        :return:

        """
        pass

    async def testRPC(self):
        rpc_name = whoami()
        LOG.info("RPC " + rpc_name + " CALLED...")
        await makeNothing()
        result = " - " + rpc_name + " result: testRPC is working!!!\n"
        LOG.info(result)
        return result



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import os
import pkg_resources

dist = pkg_resources.get_distribution(__package__)
entry_points_name = \
    os.path.join(dist.location, dist.egg_name()) + ".egg-info/entry_points.txt"

# Iotronic python package folder
package_path = os.path.join(dist.location, __package__)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"


from iotronic_lightningrod.modules.proxies import Proxy

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


import json
import os
import subprocess
import time

from iotronic_lightningrod.common.exception import NginxError
from iotronic_lightningrod.modules import utils
import iotronic_lightningrod.wampmessage as WM
from subprocess import call


class ProxyManager(Proxy.Proxy):

    def __init__(self):
        super(ProxyManager, self).__init__("nginx")

    def finalize(self):
        """Function called at the end of module loading (after RPC registration).

        :return:

        """
        pass

    def _proxyInfo(self):

        nginxMsg = {}

        try:

            get_service = 'pidof systemd > /dev/null ' \
                          '&& echo "systemd" || echo "init.d"'
            service_cmd = subprocess.Popen(get_service,
                                           shell=True, stdout=subprocess.PIPE)

            service_mng = \
                service_cmd.communicate()[0].decode("utf-8").split("\n")[0]

            if service_mng == 'init.d':
                # print('INIT')
                stat = subprocess.Popen('service nginx status',
                                        shell=True, stdout=subprocess.PIPE)
                stdout_list = stat.communicate()[0].decode("utf-8").split("\n")

                for line in stdout_list:

                    if 'running' in line:

                        nginxMsg['log'] = stdout_list[0]

                        if 'running' in line:
                            nginxMsg['status'] = True
                        else:
                            nginxMsg['status'] = False

                        nginxMsg = json.dumps(nginxMsg)

                        return nginxMsg

            elif service_mng == 'systemd':
                # print('SYSTEMD')
                stat = subprocess.Popen('systemctl status nginx.service',
                                        shell=True, stdout=subprocess.PIPE)
                stdout_list = str(stat.communicate()[0]).split('\n')
                for line in stdout_list:
                    if 'Active:' in line:

                        nginxMsg['log'] = \
                            line.split('\\n')[2].replace("   ", "")
                        if '(running)' in line:
                            nginxMsg['status'] = True
                        else:
                            nginxMsg['status'] = False

                        nginxMsg = json.dumps(nginxMsg)

                        return nginxMsg

        except Exception as err:
            LOG.error("Error check NGINX status: " + str(err))
            nginxMsg['log'] = str(err)
            nginxMsg['status'] = False
            nginxMsg = json.dumps(nginxMsg)
            return nginxMsg

    def _proxyStatus(self):

        nginxMsg = {}

        try:

            stat = subprocess.Popen(
                'systemctl status nginx.service',
                shell=True,
                stdout=subprocess.PIPE
            )
            stdout_list = str(stat.communicate()[0]).split('\n')
            for line in stdout_list:
                if 'Active:' in line:
                    if '(running)' in line:
                        nginxMsg['log'] = "NGINX is running"
                        nginxMsg['status'] = True
                        # LOG.info("--> " + nginxMsg['log'])
                    else:
                        nginxMsg['log'] = "NGINX is not running"
                        nginxMsg['status'] = False
                        # LOG.warning("--> " + nginxMsg['log'])

        except Exception as err:
            nginxMsg['log'] = "Error check NGINX status: " + str(err)
            nginxMsg['status'] = True
            # LOG.error("--> " + nginxMsg['log'])

        return json.dumps(nginxMsg)

    def _proxyReload(self):

        nginxMsg = {}

        stat = None

        try:

            stat = subprocess.call('service nginx reload', shell=True)

            if stat != 0:
                raise NginxError(str(stat))

            else:
                nginxMsg['log'] = "NGINX successfully reloaded"
                nginxMsg['code'] = stat
                LOG.info("--> " + nginxMsg['log'])

        except NginxError:
            nginxMsg['log'] = "NGINX reloading error"
            nginxMsg['code'] = stat
            LOG.warning("--> " + nginxMsg['log'])

        except Exception as err:
            nginxMsg['log'] = "NGINX Generic error: " + str(err)
            nginxMsg['code'] = stat
            LOG.warning("--> " + nginxMsg['log'])

        nginxMsg = json.dumps(nginxMsg)
        return nginxMsg

    def _proxyRestart(self):

        nginxMsg = {}

        try:
            stat = os.system('systemctl restart nginx')

            if stat != 0:
                raise NginxError(str(stat))

            else:
                nginxMsg['log'] = "NGINX successfully restart"
                nginxMsg['code'] = stat
                LOG.info("--> " + nginxMsg['log'])

        except NginxError:
            nginxMsg['log'] = "NGINX restarting error"
            nginxMsg['code'] = stat
            LOG.warning("--> " + nginxMsg['log'])

        except Exception as err:
            nginxMsg['log'] = "NGINX generic error: " + str(err)
            nginxMsg['code'] = stat
            LOG.warning("--> " + nginxMsg['log'])

        return json.dumps(nginxMsg)

    def _nginx_conf_verify(self, fp):
        with open(fp, "r") as text_file:
            LOG.debug(text_file.read())

    def _proxyEnableWebService(self, board_dns, owner_email):

        nginxMsg = {}

        try:

            nginx_path = "/etc/nginx/conf.d/"

            nginx_board_conf_file = nginx_path + "/" + board_dns + ".conf"
            nginx_board_conf = '''server {{
                listen              80;
                server_name    {0};
            }}
            '''.format(board_dns)

            with open(nginx_board_conf_file, "w") as text_file:
                text_file.write("%s" % nginx_board_conf)

            self._nginx_conf_verify(nginx_board_conf_file)
            time.sleep(3)
            self._proxyReload()
            time.sleep(3)

            command = "/usr/bin/certbot -n " \
                      "--redirect " \
                      "--authenticator webroot " \
                      "--installer nginx " \
                      "-w /var/www/html/ " \
                      "--domain " + board_dns + " " \
                      "--agree-tos " \
                      "--email " + owner_email

            LOG.debug(command)
            certbot_result = call(command, shell=True)
            LOG.info("CERTBOT RESULT: " + str(certbot_result))

            nginxMsg['result'] = "SUCCESS"
            nginxMsg['message'] = "Webservice module enabled."
            LOG.info("--> " + nginxMsg['message'])

        except Exception as err:
            nginxMsg['log'] = "NGINX DNS setup error: " + str(err)
            nginxMsg['code'] = ""
            LOG.warning("--> " + nginxMsg['log'])

        return json.dumps(nginxMsg)

    def _exposeWebservice(self, board_dns, service_dns, local_port, dns_list):

        nginxMsg = {}

        try:

            nginx_path = "/etc/nginx/conf.d"

            service_path = nginx_path + "/" + service_dns + ".conf"
            string = '''server {{
            listen              80;
            server_name         {0};

                proxy_set_header Host $http_host;
                proxy_set_header X-Real-IP $remote_addr;
                proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
                proxy_set_header X-Forwarded-Proto $scheme;
                proxy_http_version 1.1;
                proxy_set_header Upgrade $http_upgrade;
                proxy_set_header Connection "upgrade";

            location / {{
            proxy_pass http://localhost:{1};
            }}
            }}
            '''.format(service_dns, local_port)

            with open(service_path, "w") as ws_nginx_conf:
                ws_nginx_conf.write("%s" % string)

            time.sleep(3)

            self._nginx_conf_verify(service_path)

            self._proxyReload()

            time.sleep(3)

            command = "/usr/bin/certbot " \
                      "--expand -n " \
                      "--redirect " \
                      "--authenticator webroot " \
                      "--installer nginx -w /var/www/html/ " \
                      "--domain " + str(dns_list)

            command = "/usr/bin/certbot " \
                      "-n " \
                      "--redirect " \
                      "--authenticator webroot " \
                      "--installer nginx -w /var/www/html/ " \
                      "--cert-name " + str(board_dns) + " " \
                      "--domain " + str(dns_list)

            LOG.debug(command)
            certbot_result = call(command, shell=True)
            LOG.info("CERTBOT RESULT: " + str(certbot_result))

            LOG.info("Webservices list updated:\n" +
                     str(self._webserviceList()))

            nginxMsg['result'] = "SUCCESS"
            nginxMsg['message'] = "Webservice '" + service_dns + \
                                  "' exposed in NGINX."
            LOG.info(nginxMsg['message'])

        except Exception as e:
            nginxMsg['message'] = "Error exposing Webservice '" + \
                                  service_dns + \
                                  "' configuration in NGINX: {}".format(e)
            nginxMsg['result'] = "ERROR"
            LOG.warning("--> " + nginxMsg['message'])

        return json.dumps(nginxMsg)

    def _disableWebservice(self, service_dns, dns_list):
        """
        :param service:
        :param dns_list:
        :return:
        """

        nginxMsg = {}

        try:

            nginx_path = "/etc/nginx/conf.d"
            service_path = nginx_path + "/" + service_dns + ".conf"

            if os.path.exists(service_path):

                os.remove(service_path)

                time.sleep(1)

                self._proxyReload()

                time.sleep(3)

                command = "/usr/bin/certbot " \
                          "--expand -n " \
                          "--redirect " \
                          "--authenticator webroot " \
                          "--installer nginx -w /var/www/html/ " \
                          "--domain " + str(dns_list)

                LOG.debug(command)
                certbot_result = call(command, shell=True)
                LOG.info("CERTBOT RESULT: " + str(certbot_result))

                LOG.info("Webservices list updated:\n" + str(
                    self._webserviceList()))

                nginxMsg['message'] = "webservice '" \
                                      + service_dns + "' disabled."
                nginxMsg['result'] = "SUCCESS"
                LOG.info(nginxMsg['message'])

            else:
                nginxMsg['message'] = "webservice file " \
                    + service_path + " does not exist"
                nginxMsg['result'] = "ERROR"

        except Exception as e:
            nginxMsg['message'] = "Error disabling Webservice '" + \
                                  service_dns + "': {}".format(e)
            nginxMsg['result'] = "ERROR"

        return json.dumps(nginxMsg)

    def _webserviceList(self):

        nginx_path = "/etc/nginx/conf.d/"

        if os.path.exists(nginx_path):
            service_list = [f for f in os.listdir(nginx_path)
                            if os.path.isfile(os.path.join(nginx_path, f))]
        else:
            service_list = []

        return service_list

    async def NginxInfo(self):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self._proxyInfo()
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def NginxStatus(self):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self._proxyStatus()
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def NginxReload(self):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self._proxyReload()
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def NginxRestart(self):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self._proxyRestart()
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def NginxIotronicConf(self):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self._proxyIotronicConf()
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import asyncio
import inspect
import pkg_resources
from six import moves
from stevedore import extension
import sys

from iotronic_lightningrod.config import entry_points_name
from iotronic_lightningrod.lightningrod import SESSION
from iotronic_lightningrod.modules import Module


from oslo_log import log as logging
LOG = logging.getLogger(__name__)


def getFuncName():
    return inspect.stack()[1][3]


def refresh_stevedore(namespace=None):
    """Trigger reload of entry points.

    Useful to have dynamic loading/unloading of stevedore modules.
    """
    # NOTE(sheeprine): pkg_resources doesn't support reload on python3 due to
    # defining basestring which is still there on reload hence executing
    # python2 related code.
    try:
        del sys.modules['pkg_resources'].basestring
    except AttributeError:
        # python2, do nothing
        pass
    # Force working_set reload
    moves.reload_module(sys.modules['pkg_resources'])
    # Clear stevedore cache
    cache = extension.ExtensionManager.ENTRY_POINT_CACHE
    if namespace:
        if namespace in cache:
            del cache[namespace]
    else:
        cache.clear()


class Utility(Module.Module):

    def __init__(self, board, session):
        super(Utility, self).__init__("Utility", board)

    def finalize(self):
        pass

    def restore(self):
        pass

    async def hello(self, client_name, message):
        import random
        s = random.uniform(0.5, 3.0)
        await asyncio.sleep(s)
        result = "Hello by board to Conductor " + client_name + \
                 " that said me " + message + " - Time: " + '%.2f' % s
        LOG.info("DEVICE hello result: " + str(result))

        return result

    async def plug_and_play(self, new_module, new_class):
        LOG.info("LR modules loaded:\n\t" + new_module)

        # Updating entry_points
        with open(entry_points_name, 'a') as entry_points:
            entry_points.write(
                new_module +
                '= iotronic_lightningrod.modules.' + new_module + ':'
                + new_class
            )

            # Reload entry_points
            refresh_stevedore('s4t.modules')
            LOG.info("New entry_points loaded!")

        # Reading updated entry_points
        named_objects = {}
        for ep in pkg_resources.iter_entry_points(group='s4t.modules'):
            named_objects.update({ep.name: ep.load()})

        await named_objects

        SESSION.disconnect()

        return str(named_objects)

    async def changeConf(self, conf):

        await self.board.getConf(conf)

        self.board.setUpdateTime()

        result = "Board configuration changed!"
        LOG.info("PROVISIONING RESULT: " + str(result))

        return result

    async def destroyNode(self, conf):

        await self.board.setConf(conf)

        result = "Board configuration cleaned!"
        LOG.info("DESTROY RESULT: " + str(result))

        return result



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import signal

from oslo_log import log as logging
LOG = logging.getLogger(__name__)

from iotronic_lightningrod.common import utils


def manageTimeout(error_message, action):
    try:

        raise TimeoutError(error_message, action)

    except TimeoutError as err:
        details = err.args[0]
        if (action == "ws_alive"):

            LOG.warning("Iotronic RPC-ALIVE timeout details: " + str(details))
            try:

                utils.destroyWampSocket()

            except Exception as e:
                LOG.warning("Iotronic RPC-ALIVE timeout error: " + str(e))

        else:
            LOG.warning("Board connection call timeout ["
                        + str(action) + "]: " + str(details))
            utils.LR_restart()


class NginxError(Exception):

    def __init__(self, message):
        super(NginxError, self).__init__(message)


class TimeoutError(Exception):

    def __init__(self, message, action):
        super(TimeoutError, self).__init__(message)

        self.action = action


class timeout(object):

    def __init__(self, seconds=1, error_message='Timeout', action=None):
        self.seconds = seconds
        self.error_message = error_message
        self.action = action

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message, self.action)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class timeoutRPC(object):

    def __init__(self, seconds=1, error_message='Timeout-RPC', action=None):
        self.seconds = seconds
        self.error_message = error_message
        self.action = action

    def handle_timeout(self, signum, frame):
        manageTimeout(self.error_message, self.action)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class timeoutALIVE(object):

    def __init__(self, seconds=1, error_message='Timeout-Alive', action=None):
        self.seconds = seconds
        self.error_message = error_message
        self.action = action

    def handle_timeout(self, signum, frame):
        manageTimeout(self.error_message, self.action)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)



# Copyright 2018 MDSLAB - University of Messina
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"


from oslo_log import log as logging
LOG = logging.getLogger(__name__)


import os
import pkg_resources
import psutil
import subprocess
import sys


def LR_restart():
    try:
        LOG.warning("Lightning-rod RESTARTING...")
        python = sys.executable
        os.execl(python, python, *sys.argv)
    except Exception as err:
        LOG.error("Lightning-rod restarting error" + str(err))


def checkIotronicConf(lr_CONF):

    try:
        if(lr_CONF.log_file == None):
            LOG.warning("'log_file' is not specified!")
            return False
        else:
            print("View logs in " + lr_CONF.log_file)
            return True
    except Exception as err:
        print(err)
        return False


def destroyWampSocket():

    LR_PID = os.getpid()

    try:
        process = subprocess.Popen(
            ["gdb", "-p", str(LR_PID)],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE
        )

        proc = psutil.Process()
        print("WAMP RECOVERY: " + str(proc.connections()[0]))

        ws_fd = proc.connections()[0].fd
        first = b"call ((void(*)()) shutdown)("
        fd = str(ws_fd).encode('ascii')
        last = b"u,0)\nquit\ny"
        commands = b"%s%s%s" % (first, fd, last)
        process.communicate(input=commands)[0]

        msg = "Websocket-Zombie closed! Restoring..."
        LOG.warning(msg)
        print(msg)

    except Exception as e:
        LOG.warning("RPC-ALIVE - destroyWampSocket error: " + str(e))
        LR_restart()


def get_version(package):
    package = package.lower()
    return next((p.version for p in pkg_resources.working_set if
                 p.project_name.lower() == package), "No version")



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

from datetime import datetime
import json
import os

from oslo_config import cfg
from oslo_log import log as logging
LOG = logging.getLogger(__name__)

CONF = cfg.CONF

SETTINGS = '/etc/iotronic/settings.json'


class Board(object):

    def __init__(self):
        self.iotronic_config = {}

        self.board_config = {}
        self.name = None
        self.type = None
        self.status = None
        self.uuid = None
        self.code = None
        self.agent = None
        self.mobile = None
        self.session = None
        self.session_id = None
        self.agent_url = None

        self.location = {}

        self.device = None
        self.proxy = None

        self.wamp_config = None
        self.extra = {}

        self.loadSettings()

    def loadConf(self):
        """This method loads the JSON configuraton file: settings.json.

        :return:

        """

        try:

            with open(SETTINGS) as settings:
                lr_settings = json.load(settings)

        except Exception as err:
            LOG.error("Parsing error in " + SETTINGS + ": " + str(err))
            lr_settings = None

        return lr_settings

    def loadSettings(self):
        '''This method gets and sets the board attributes from the conf file.

        '''

        # Load all settings.json file
        self.iotronic_config = self.loadConf()

        try:
            # STATUS OPERATIVE
            board_config = self.iotronic_config['iotronic']['board']
            self.uuid = board_config['uuid']
            self.code = board_config['code']
            self.name = board_config['name']
            self.status = board_config['status']
            self.type = board_config['type']
            self.mobile = board_config['mobile']
            self.extra = board_config['extra']
            self.agent = board_config['agent']
            self.created_at = board_config['created_at']
            self.updated_at = board_config['updated_at']  # self.getTimestamp()
            self.location = board_config['location']

            self.extra = self.iotronic_config['iotronic']['extra']

            LOG.info('Board settings:')
            LOG.info(' - code: ' + str(self.code))
            LOG.info(' - uuid: ' + str(self.uuid))
            # LOG.debug(" - conf:\n" + json.dumps(board_config, indent=4))

            self.getWampAgent(self.iotronic_config)

        except Exception as err:
            if str(err) != 'uuid':
                LOG.warning("settings.json file exception: " + str(err))

            # STATUS REGISTERED
            try:
                self.code = board_config['code']
                LOG.info('First registration board settings: ')
                LOG.info(' - code: ' + str(self.code))
                self.getWampAgent(self.iotronic_config)
            except Exception as err:
                LOG.error("Wrong code: " + str(err))
                os._exit(1)

    def getWampAgent(self, config):
        '''This method gets and sets the WAMP Board attributes from the conf file.

        '''
        try:

            self.wamp_config = config['iotronic']['wamp']['main-agent']
            LOG.info('WAMP Agent settings:')

        except Exception:
            if (self.status is None) | (self.status == "registered"):
                self.wamp_config = \
                    config['iotronic']['wamp']['registration-agent']
                LOG.info('Registration Agent settings:')
            else:
                LOG.error(
                    "WAMP Agent configuration is wrong... "
                    "please check settings.json WAMP configuration... Bye!"
                )
                os._exit(1)

        # self.agent_url = str(self.wamp_config['url'])
        LOG.info(' - agent: ' + str(self.agent))
        LOG.info(' - url: ' + str(self.wamp_config['url']))
        LOG.info(' - realm: ' + str(self.wamp_config['realm']))
        # LOG.debug("- conf:\n" + json.dumps(self.wamp_config, indent=4))

    def setConf(self, conf):
        # LOG.info("\nNEW CONFIGURATION:\n" + str(json.dumps(conf, indent=4)))

        with open(SETTINGS, 'w') as f:
            json.dump(conf, f, indent=4)

        # Reload configuration
        self.loadSettings()

    def updateStatus(self, status):
        self.iotronic_config['iotronic']['board']["status"] = status

        with open(SETTINGS, 'w') as f:
            json.dump(self.iotronic_config, f, indent=4)

    def getTimestamp(self):
        # datetime.now(tzlocal()).isoformat()
        return datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')

    def setUpdateTime(self):
        self.iotronic_config['iotronic']['board']["updated_at"] = \
            self.updated_at

        with open(SETTINGS, 'w') as f:
            json.dump(self.iotronic_config, f, indent=4)



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

from iotronic_lightningrod.config import package_path
from iotronic_lightningrod.lightningrod import RPC_proxies
from iotronic_lightningrod.lightningrod import SESSION
from iotronic_lightningrod.modules import Module
from iotronic_lightningrod.modules import utils
import iotronic_lightningrod.wampmessage as WM


import importlib as imp
import inspect
import json
import OpenSSL.crypto
import os
import time

from oslo_config import cfg
from oslo_log import log as logging
LOG = logging.getLogger(__name__)

proxy_opts = [
    cfg.StrOpt(
        'proxy',
        choices=[('nginx', ('nginx proxy')), ],
        help=('Proxy for WebServices Manager')
    ),
]

CONF = cfg.CONF

webservice_group = cfg.OptGroup(
    name='webservices', title='WebServices options'
)
CONF.register_group(webservice_group)
CONF.register_opts(proxy_opts, group=webservice_group)


class WebServiceManager(Module.Module):

    def __init__(self, board, session):
        super(WebServiceManager, self).__init__("WebServiceManager", board)

        LOG.info(" - Proxy used: " + CONF.webservices.proxy.upper())

        try:
            proxy_type = CONF.webservices.proxy
            path = package_path + "/modules/proxies/" + proxy_type + ".py"

            if os.path.exists(path):

                proxy_module = imp.import_module(
                    "iotronic_lightningrod.modules.proxies." + proxy_type
                )

                LOG.info(" --> " + proxy_type.upper() + " module imported!")

                proxy = proxy_module.ProxyManager()

                proxy_meth_list = inspect.getmembers(
                    proxy,
                    predicate=inspect.ismethod
                )

                RPC_proxies[proxy_type] = proxy_meth_list

                board.proxy = proxy

                self._proxyWampRegister(proxy_meth_list, board)

            else:
                LOG.warning("Proxy '" + proxy_type + "' not supported!")

        except Exception as err:
            LOG.error("Error init WebServiceManager: " + str(err))

    def finalize(self):

        try:

            proxy_status = json.loads(self.board.proxy._proxyInfo())
            LOG.info("--> Proxy " + self.board.proxy.type.upper()
                     + " status:\n Active: " + str(proxy_status['status'])
                     + "\n Info: " + str(proxy_status['log']))

            LOG.info("Webservice exposed on device:")
            active_webservice_list = self.board.proxy._webserviceList()
            if len(active_webservice_list) != 0:
                for ws in active_webservice_list:
                    ws = ws.replace('.conf', '')
                    LOG.info("-> " + ws)
            else:
                LOG.info("-> NO WebService!")

            LOG.info("Certificates on device:")
            active_certs_list = self._certsList()
            if len(active_certs_list) != 0:
                for certificate in active_certs_list:
                    LOG.info("-> " + certificate)

                    c = open('/etc/letsencrypt/live/'
                             + certificate + '/cert.pem').read()
                    cert = OpenSSL.crypto.load_certificate(
                        OpenSSL.crypto.FILETYPE_PEM, c)

                    LOG.info("--> Subject: " + str(cert.get_subject()))
                    LOG.info("--> ISSUER Organization:" +
                             str(cert.get_issuer()))
                    LOG.info("--> Expire date: " + str(
                        cert.get_notAfter().decode("utf-8")))
                    LOG.info("--> Expired: " + str(cert.has_expired()))

            else:
                LOG.info("-> NO certificates!")

            # Safe apply changes on proxy
            self.board.proxy._proxyReload()
            time.sleep(3)

            LOG.info("WebService Manager initialized!")

        except Exception as err:
            LOG.error("Error finalize init WebServiceManager: " + str(err))

    def restore(self):
        LOG.info("WebService Manager restored.")

    def _certsList(self):

        letsencrypt_path = "/etc/letsencrypt/live/"

        if os.path.exists(letsencrypt_path):
            certs_list = [
                f for f in os.listdir(letsencrypt_path)
                if os.path.isdir(os.path.join(letsencrypt_path, f))
            ]
        else:
            certs_list = []

        return certs_list

    def _proxyWampRegister(self, proxy_meth_list, board):

        LOG.info(" - " + str(board.proxy.type).upper()
                 + " proxy registering RPCs:")

        for meth in proxy_meth_list:
            if (meth[0] != "__init__") & (meth[0] != "finalize") \
                    & (meth[0] != "restore"):
                # LOG.info(" - " + str(meth[0]))
                rpc_addr = u'iotronic.' + str(board.session_id) + '.' + \
                           board.uuid + '.' + meth[0]

                # LOG.debug(" --> " + str(rpc_addr))
                if not meth[0].startswith('_'):
                    SESSION.register(meth[1], rpc_addr)
                    LOG.info("   --> " + str(meth[0]))

    async def ExposeWebservice(self, board_dns, service_dns,
                               local_port, dns_list):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        response = self.board.proxy._exposeWebservice(board_dns, service_dns,
                                                      local_port, dns_list)

        response = json.loads(response)

        if(response['result'] == "SUCCESS"):
            message = "Webservice '" + service_dns + "' successfully exposed!"
            LOG.info("--> " + str(message))
            w_msg = WM.WampSuccess(response)
        else:
            LOG.warning("--> " + str(response['message']))
            w_msg = WM.WampWarning(response)

        return w_msg.serialize()

    async def UnexposeWebservice(self, service, dns_list):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        response = self.board.proxy._disableWebservice(service, dns_list)

        response = json.loads(response)

        if (response['result'] == "SUCCESS"):
            LOG.info("--> " + str(response['message']))
            w_msg = WM.WampSuccess(response)
        else:
            LOG.warning("--> " + str(response['message']))
            w_msg = WM.WampWarning(response)

        return w_msg.serialize()

    async def EnableWebService(self, board_dns, owner_email):

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = self.board.proxy._proxyEnableWebService(
            board_dns,
            owner_email
        )
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import json

SUCCESS = 'SUCCESS'
ERROR = 'ERROR'
WARNING = 'WARNING'


def deserialize(received):
    m = json.loads(received)
    return WampMessage(**m)


class WampMessage(object):
    def __init__(self, message=None, result=None):
        self.message = message
        self.result = result

    def serialize(self):
        return json.dumps(self, default=lambda o: o.__dict__)
    """
    def deserialize(self, received):
        self.__dict__ = json.loads(received)
        return self
    """


class WampSuccess(WampMessage):
    def __init__(self, msg=None):
        super(WampSuccess, self).__init__(msg, SUCCESS)


class WampError(WampMessage):
    def __init__(self, msg=None):
        super(WampError, self).__init__(msg, ERROR)


class WampWarning(WampMessage):
    def __init__(self, msg=None):
        super(WampWarning, self).__init__(msg, WARNING)



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

from datetime import datetime
import importlib as imp
import json
import os
import queue
import shutil
import time


from iotronic_lightningrod.modules import Module
from iotronic_lightningrod.modules.plugins import PluginSerializer
from iotronic_lightningrod.modules import utils
import iotronic_lightningrod.wampmessage as WM


from oslo_config import cfg
from oslo_log import log as logging
LOG = logging.getLogger(__name__)

CONF = cfg.CONF
PLUGINS_THRS = {}
PLUGINS_CONF_FILE = CONF.lightningrod_home + "/plugins.json"


class PluginManager(Module.Module):

    """Plugin module to manage board plugins.

    """

    def __init__(self, board, session):
        """Init function for PluginManager module.

        :param board:
        :param session:

        """

        # Module declaration
        super(PluginManager, self).__init__("PluginManager", board)

        # Creation of plugins.json configuration file
        self._createPluginsConf()

    def finalize(self):
        """Function called at the end of module loading.

        This function in this module reloads
        the enabled (asynchronous) plugins at boot.

        """

        # Reboot boot enabled plugins
        self._rebootOnBootPlugins()

    def restore(self):
        pass

    def _loadPluginsConf(self):
        """Load plugins.json JSON configuration.

        :return: JSON Plugins configuration

        """

        try:

            with open(PLUGINS_CONF_FILE) as settings:
                plugins_conf = json.load(settings)

        except Exception as err:
            LOG.error(
                "Parsing error in " + PLUGINS_CONF_FILE + ": " + str(err))
            plugins_conf = None

        return plugins_conf

    def _getEnabledPlugins(self):
        """This function gets the list of all asynchronous plugins.

        We considered only those plugins with 'callable' flag set to False
        and 'onboot' flag set to True.

        :return: enabledPlugins List

        """
        enabledPlugins = []
        plugins_conf = self._loadPluginsConf()

        for plugin in plugins_conf['plugins']:

            if plugins_conf['plugins'][plugin]['callable'] is False:

                if plugins_conf['plugins'][plugin]['onboot'] is True:

                    if plugins_conf['plugins'][plugin]['status'] == \
                            "operative":
                        enabledPlugins.append(plugin)

        if len(enabledPlugins) != 0:
            LOG.info(" - Enabled plugins list: " + str(enabledPlugins))

        return enabledPlugins

    def _rebootOnBootPlugins(self):
        """Reboot at boot each enabled asynchronous plugin

        :return:

        """

        f_name = utils.getFuncName()
        LOG.info("Rebooting enabled plugins:")

        enabledPlugins = self._getEnabledPlugins()

        if enabledPlugins.__len__() == 0:

            message = "No plugin to reboot!"
            LOG.info(" - " + message)

        else:

            for plugin_uuid in enabledPlugins:

                plugins_conf = self._loadPluginsConf()
                plugin_name = plugins_conf['plugins'][plugin_uuid]['name']

                try:

                    if (plugin_uuid in PLUGINS_THRS) and (
                            PLUGINS_THRS[plugin_uuid].isAlive()
                    ):

                        LOG.warning(" - Plugin "
                                    + plugin_uuid + " already started!")

                    else:

                        LOG.info(" - Rebooting plugin " + plugin_uuid)

                        plugin_home = \
                            CONF.lightningrod_home + "/plugins/" + plugin_uuid
                        plugin_filename = \
                            plugin_home + "/" + plugin_uuid + ".py"
                        plugin_params_file = \
                            plugin_home + "/" + plugin_uuid + ".json"

                        if os.path.exists(plugin_filename):

                            task = imp.machinery.SourceFileLoader(
                                "plugin",
                                plugin_filename
                            ).load_module()

                            if os.path.exists(plugin_params_file):

                                with open(plugin_params_file) as conf:
                                    plugin_params = json.load(conf)

                                worker = task.Worker(
                                    plugin_uuid,
                                    plugin_name,
                                    q_result=None,
                                    params=plugin_params
                                )

                                PLUGINS_THRS[plugin_uuid] = worker
                                LOG.info("   - Starting plugin " + str(worker))

                                worker.start()

                            else:
                                message = "ERROR " + plugin_params_file \
                                          + " does not exist!"

                                LOG.error("   - "
                                          + worker.complete(f_name, message))

                        else:
                            message = "ERROR " \
                                      + plugin_filename + " does not exist!"

                            LOG.error(
                                "   - " + worker.complete(f_name, message))

                    message = "rebooted!"

                    LOG.info("   - " + worker.complete(f_name, message))

                except Exception as err:
                    message = "Error rebooting plugin " \
                              + plugin_uuid + ": " + str(err)
                    LOG.error(" - " + message)

    def _createPluginsConf(self):
        """Create plugins.json file if it does not exist.

        """
        if not os.path.exists(PLUGINS_CONF_FILE):
            LOG.debug("plugins.json does not exist: creating...")
            plugins_conf = {'plugins': {}}
            with open(PLUGINS_CONF_FILE, 'w') as f:
                json.dump(plugins_conf, f, indent=4)

    async def PluginInject(self, plugin, onboot):
        """Plugin injection procedure into the board:

         1. get Plugin files
         2. deserialize files
         3. store files

        :param plugin:
        :param onboot:
        :return:

        """

        rpc_name = utils.getFuncName()

        try:

            plugin_uuid = plugin['uuid']
            plugin_name = plugin['name']
            code = plugin['code']
            callable = plugin['callable']

            LOG.info("RPC " + rpc_name + " for plugin '"
                     + plugin_name + "' (" + plugin_uuid + ")")

            # Deserialize the plugin code received
            ser = PluginSerializer.ObjectSerializer()
            loaded = ser.deserialize_entity(code)
            # LOG.debug("- plugin loaded code:\n" + loaded)

            plugin_path = CONF.lightningrod_home \
                + "/plugins/" + plugin_uuid + "/"
            plugin_filename = plugin_path + plugin_uuid + ".py"

            # Plugin folder creation if does not exist
            if not os.path.exists(plugin_path):
                os.makedirs(plugin_path)

            # Plugin code file creation
            with open(plugin_filename, "w") as pluginfile:
                pluginfile.write(loaded)

            # Load plugins.json configuration file
            plugins_conf = self._loadPluginsConf()

            # LOG.debug("Plugin setup:\n"
            #          + json.dumps(plugin, indent=4, sort_keys=True))

            # Save plugin settings in plugins.json
            if plugin_uuid not in plugins_conf['plugins']:

                # It is a new plugin
                plugins_conf['plugins'][plugin_uuid] = {}
                plugins_conf['plugins'][plugin_uuid]['name'] = plugin_name
                plugins_conf['plugins'][plugin_uuid]['onboot'] = onboot
                plugins_conf['plugins'][plugin_uuid]['callable'] = callable
                plugins_conf['plugins'][plugin_uuid]['injected_at'] = \
                    datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
                plugins_conf['plugins'][plugin_uuid]['updated_at'] = ""
                plugins_conf['plugins'][plugin_uuid]['status'] = "injected"

                LOG.info(" - Plugin '" + plugin_name + "' created!")
                message = rpc_name + " result: INJECTED"

            else:
                # The plugin was already injected and we are updating it
                plugins_conf['plugins'][plugin_uuid]['name'] = plugin_name
                plugins_conf['plugins'][plugin_uuid]['onboot'] = onboot
                plugins_conf['plugins'][plugin_uuid]['callable'] = callable
                plugins_conf['plugins'][plugin_uuid]['updated_at'] = \
                    datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
                plugins_conf['plugins'][plugin_uuid]['status'] = "updated"

                LOG.info("Plugin '" + plugin_name
                         + "' (" + str(plugin_uuid) + ") updated!")
                message = rpc_name + " result: UPDATED"

            LOG.info(" - Plugin setup:\n" + json.dumps(
                plugins_conf['plugins'][plugin_uuid],
                indent=4,
                sort_keys=True
            ))

            # Apply the changes to plugins.json
            with open(PLUGINS_CONF_FILE, 'w') as f:
                json.dump(plugins_conf, f, indent=4)

            LOG.info(" - " + message)
            w_msg = WM.WampSuccess(message)

        except Exception as err:
            message = "Plugin injection error: " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()

    async def PluginStart(self, plugin_uuid, parameters=None):
        """To start an asynchronous plugin;

        the plugin will run until the PluginStop is called.

        :param plugin_uuid:
        :param parameters:
        :return: return a response to RPC request

        """

        try:

            rpc_name = utils.getFuncName()
            LOG.info("RPC " + rpc_name + " called for '"
                     + plugin_uuid + "' plugin:")

            plugins_conf = self._loadPluginsConf()

            if plugin_uuid in plugins_conf['plugins']:

                plugin_name = plugins_conf['plugins'][plugin_uuid]['name']

                # Check if the plugin is already running
                if (plugin_uuid in PLUGINS_THRS) and (
                        PLUGINS_THRS[plugin_uuid].isAlive()
                ):

                    message = "ALREADY STARTED!"
                    LOG.warning(" - Plugin "
                                + plugin_uuid + " already started!")
                    w_msg = WM.WampError(message)

                else:

                    plugin_home = \
                        CONF.lightningrod_home + "/plugins/" + plugin_uuid
                    plugin_filename = \
                        plugin_home + "/" + plugin_uuid + ".py"
                    plugin_params_file = \
                        plugin_home + "/" + plugin_uuid + ".json"

                    # Import plugin (as python module)
                    if os.path.exists(plugin_filename):

                        task = imp.machinery.SourceFileLoader(
                            "plugin",
                            plugin_filename
                        ).load_module()

                        LOG.info(" - Plugin '" + plugin_uuid + "' imported!")

                        # Store input parameters of the plugin
                        if parameters is not None:

                            with open(plugin_params_file, 'w') as f:
                                json.dump(parameters, f, indent=4)

                            with open(plugin_params_file) as conf:
                                plugin_params = json.load(conf)

                            LOG.info(" - plugin with parameters:")
                            LOG.info("   " + str(plugin_params))

                        else:
                            plugin_params = None

                        worker = task.Worker(
                            plugin_uuid,
                            plugin_name,
                            params=plugin_params
                        )

                        PLUGINS_THRS[plugin_uuid] = worker
                        LOG.debug(" - Starting plugin " + str(worker))

                        worker.start()

                        # Apply the changes to plugins.json
                        with open(PLUGINS_CONF_FILE, 'w') as f:
                            plugins_conf['plugins'][plugin_uuid]['status'] = \
                                'operative'
                            json.dump(plugins_conf, f, indent=4)

                        response = "STARTED"
                        LOG.info(" - " + worker.complete(rpc_name, response))
                        w_msg = WM.WampSuccess(response)

                    else:
                        message = \
                            rpc_name + " - ERROR " \
                            + plugin_filename + " does not exist!"
                        LOG.error(" - " + message)
                        w_msg = WM.WampError(message)

            else:
                message = "Plugin " + plugin_uuid \
                          + " does not exist in this board!"
                LOG.warning(" - " + message)
                w_msg = WM.WampError(message)

        except Exception as err:
            message = \
                rpc_name + " - ERROR - plugin (" + plugin_uuid + ") - " \
                + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()

    async def PluginStop(self, plugin_uuid, parameters=None):
        """To stop an asynchronous plugin

        :param plugin_uuid: ID of plufin to stop
        :param parameters: JSON OPTIONAL stop parameters; 'delay' in seconds
        :return: return a response to RPC request

        """
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED for '"
                 + plugin_uuid + "' plugin:")

        if parameters is not None:
            LOG.info(" - " + rpc_name + " parameters: " + str(parameters))
            if 'delay' in parameters:
                delay = parameters['delay']
                LOG.info(" --> stop delay: " + str(delay))

        try:

            if plugin_uuid in PLUGINS_THRS:

                worker = PLUGINS_THRS[plugin_uuid]
                LOG.debug(" - Stopping plugin " + str(worker))

                if worker.isAlive():

                    if 'delay' in parameters:
                        time.sleep(delay)

                    worker.stop()

                    del PLUGINS_THRS[plugin_uuid]

                    message = "STOPPED"
                    LOG.info(" - " + worker.complete(rpc_name, message))
                    w_msg = WM.WampSuccess(message)

                else:
                    message = \
                        rpc_name \
                        + " - ERROR - plugin (" + plugin_uuid \
                        + ") is instantiated but is not running anymore!"
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

            else:
                message = \
                    rpc_name + " - WARNING " \
                    + plugin_uuid + "  is not running!"
                LOG.warning(" - " + message)
                w_msg = WM.WampWarning(message)

        except Exception as err:
            message = \
                rpc_name \
                + " - ERROR - plugin (" + plugin_uuid + ") - " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()

    async def PluginCall(self, plugin_uuid, parameters=None):
        """To execute a synchronous plugin into the board

        :param plugin_uuid:
        :param parameters:
        :return: return a response to RPC request

        """

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED for " + plugin_uuid + " plugin:")

        try:

            if (plugin_uuid in PLUGINS_THRS) and (
                    PLUGINS_THRS[plugin_uuid].isAlive()
            ):

                message = "Plugin " + plugin_uuid + " already started!"
                LOG.warning(" - " + message)
                w_msg = WM.WampWarning(message)

            else:

                plugin_home = CONF.lightningrod_home \
                    + "/plugins/" + plugin_uuid
                plugin_filename = plugin_home + "/" + plugin_uuid + ".py"
                plugin_params_file = plugin_home + "/" + plugin_uuid + ".json"

                plugins_conf = self._loadPluginsConf()
                plugin_name = plugins_conf['plugins'][plugin_uuid]['name']

                # Import plugin (as python module)
                if os.path.exists(plugin_filename):

                    try:

                        # task = imp.load_source("plugin", plugin_filename)
                        task = imp.machinery.SourceFileLoader(
                            "plugin",
                            plugin_filename
                        ).load_module()

                        LOG.info(" - Plugin " + plugin_uuid + " imported!")

                        q_result = queue.Queue()

                    except Exception as err:
                        message = "Error importing plugin " \
                                  + plugin_filename + ": " + str(err)
                        LOG.error(" - " + message)
                        w_msg = WM.WampError(str(err))

                        return w_msg.serialize()

                    try:

                        # Store input parameters of the plugin
                        if parameters is not None:
                            with open(plugin_params_file, 'w') as f:
                                json.dump(parameters, f, indent=4)

                            with open(plugin_params_file) as conf:
                                plugin_params = json.load(conf)

                            LOG.info(" - Plugin configuration:\n"
                                     + str(plugin_params))

                        else:
                            plugin_params = None

                        worker = task.Worker(
                            plugin_uuid,
                            plugin_name,
                            q_result=q_result,
                            params=plugin_params
                        )

                        PLUGINS_THRS[plugin_uuid] = worker
                        LOG.debug(" - Executing plugin " + str(worker))

                        worker.start()

                        while q_result.empty():
                            pass

                        response = q_result.get()

                        LOG.info(" - " + worker.complete(rpc_name, response))
                        w_msg = WM.WampSuccess(response)

                    except Exception as err:
                        message = "Error spawning plugin " \
                                  + plugin_filename + ": " + str(err)
                        LOG.error(" - " + message)
                        w_msg = WM.WampError(str(err))

                        return w_msg.serialize()

                else:
                    message = \
                        rpc_name \
                        + " - ERROR " + plugin_filename + " does not exist!"
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

        except Exception as err:
            message = \
                rpc_name \
                + " - ERROR - plugin (" + plugin_uuid + ") - " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()

    async def PluginRemove(self, plugin_uuid):
        """To remove a plugin from the board

        :param plugin_uuid:
        :return: return a response to RPC request

        """

        rpc_name = utils.getFuncName()

        LOG.info("RPC " + rpc_name + " for plugin " + plugin_uuid)

        plugin_path = CONF.lightningrod_home + "/plugins/" + plugin_uuid + "/"

        if os.path.exists(plugin_path) is False \
                or os.path.exists(PLUGINS_CONF_FILE) is False:

            message = "Plugin paths or files do not exist!"
            LOG.error(message)
            w_msg = WM.WampError(message)

            return w_msg.serialize()

        else:

            LOG.info(" - Removing plugin...")

            try:

                try:

                    shutil.rmtree(
                        plugin_path,
                        ignore_errors=False,
                        onerror=None
                    )

                except Exception as err:
                    message = "Removing plugin's files error in " \
                              + plugin_path + ": " + str(err)
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(str(err))

                    return w_msg.serialize()

                # Remove from plugins.json file its configuration
                try:

                    plugins_conf = self._loadPluginsConf()

                    if plugin_uuid in plugins_conf['plugins']:

                        plugin_name = \
                            plugins_conf['plugins'][plugin_uuid]['name']

                        del plugins_conf['plugins'][plugin_uuid]

                        with open(PLUGINS_CONF_FILE, 'w') as f:
                            json.dump(plugins_conf, f, indent=4)

                        if plugin_uuid in PLUGINS_THRS:
                            worker = PLUGINS_THRS[plugin_uuid]
                            if worker.isAlive():
                                LOG.info(" - Plugin '"
                                         + plugin_name + "' is running...")
                                worker.stop()
                                LOG.info("   ...stopped!")

                            del PLUGINS_THRS[plugin_uuid]

                        message = "PluginRemove result: " \
                                  + plugin_uuid + " removed!"
                        LOG.info(" - " + message)

                    else:
                        message = "PluginRemove result:  " \
                                  + plugin_uuid + " already removed!"
                        LOG.warning(" - " + message)

                    w_msg = WM.WampSuccess(message)

                    return w_msg.serialize()

                except Exception as err:
                    message = "Updating plugins.json error: " + str(err)
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(str(err))

                    return w_msg.serialize()

            except Exception as err:
                message = "Plugin removing error: {0}".format(err)
                LOG.error(" - " + message)
                w_msg = WM.WampError(str(err))

                return w_msg.serialize()

    async def PluginReboot(self, plugin_uuid, parameters=None):
        """To reboot an asynchronous plugin (callable = false) into the board.

        :return: return a response to RPC request

        """

        rpc_name = utils.getFuncName()

        LOG.info("RPC " + rpc_name + " CALLED for '"
                 + plugin_uuid + "' plugin:")

        LOG.info(" - plugin restarting with parameters:")
        LOG.info("   " + str(parameters))

        try:

            plugin_home = CONF.lightningrod_home + "/plugins/" + plugin_uuid
            plugin_filename = plugin_home + "/" + plugin_uuid + ".py"
            plugin_params_file = plugin_home + "/" + plugin_uuid + ".json"

            plugins_conf = self._loadPluginsConf()
            plugin_name = plugins_conf['plugins'][plugin_uuid]['name']
            callable = plugins_conf['plugins'][plugin_uuid]['callable']

            if callable is False:

                if plugin_uuid in PLUGINS_THRS:

                    worker = PLUGINS_THRS[plugin_uuid]

                    # STOP PLUGIN----------------------------------------------

                    if worker.isAlive():

                        LOG.info(" - Thread "
                                 + plugin_uuid + " is running, stopping...")
                        LOG.debug(" - Stopping plugin " + str(worker))
                        worker.stop()

                    while worker.isAlive():
                        pass

                    # Remove from plugin thread list
                    del PLUGINS_THRS[plugin_uuid]
                    LOG.debug(" - plugin data structure cleaned!")

                # START PLUGIN-------------------------------------------------
                if os.path.exists(plugin_filename):

                    # Import plugin python module
                    # task = imp.load_source("plugin", plugin_filename)
                    task = imp.machinery.SourceFileLoader(
                        "plugin",
                        plugin_filename
                    ).load_module()

                    if parameters is None:

                        if os.path.exists(plugin_params_file):

                            with open(plugin_params_file) as conf:
                                plugin_params = json.load(conf)

                        else:
                            plugin_params = None

                    else:
                        plugin_params = parameters
                        LOG.info(" - plugin restarting with parameters:")
                        LOG.info("   " + str(plugin_params))

                    worker = task.Worker(
                        plugin_uuid,
                        plugin_name,
                        params=plugin_params
                    )

                    PLUGINS_THRS[plugin_uuid] = worker
                    LOG.info("   - Starting plugin " + str(worker))

                    worker.start()

                    message = "REBOOTED"
                    LOG.info(" - " + worker.complete(rpc_name, message))
                    w_msg = WM.WampSuccess(message)

                else:
                    message = "ERROR '" + plugin_filename + "' does not exist!"
                    LOG.error(" - " + message)
                    w_msg = WM.WampError(message)

        except Exception as err:
            message = "Error rebooting plugin '" \
                      + plugin_uuid + "': " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()

    async def PluginStatus(self, plugin_uuid):
        """Check status thread plugin

        :param plugin_uuid:
        :return:

        """

        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED for '"
                 + plugin_uuid + "' plugin:")

        try:

            if plugin_uuid in PLUGINS_THRS:

                worker = PLUGINS_THRS[plugin_uuid]

                if worker.isAlive():
                    result = "ALIVE"
                else:
                    result = "DEAD"

                LOG.info(" - " + worker.complete(rpc_name, result))
                w_msg = WM.WampSuccess(result)

            else:
                result = "DEAD"
                LOG.info(" - " + rpc_name + " result for "
                         + plugin_uuid + ": " + result)
                w_msg = WM.WampSuccess(result)

        except Exception as err:
            message = \
                rpc_name \
                + " - ERROR - plugin (" + plugin_uuid + ") - " + str(err)
            LOG.error(" - " + message)
            w_msg = WM.WampError(str(err))

            return w_msg.serialize()

        return w_msg.serialize()



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import abc
import six

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class Device(object):
    """Base class for each s4t Lightning-rod device.

    """

    def __init__(self, device_type):
        self.device_type = device_type

    def finalize(self):
        pass



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

from oslo_log import log as logging

from iotronic_lightningrod.devices.gpio import Gpio

LOG = logging.getLogger(__name__)


class RaspberryGpio(Gpio.Gpio):
    def __init__(self):
        super(RaspberryGpio, self).__init__("raspberry")
        LOG.info("Raspberry GPIO module importing...")

    # Enable GPIO
    def EnableGPIO(self):
        result = " - GPIO not available for 'raspberry' device!"
        LOG.info(result)

    def DisableGPIO(self):
        result = " - GPIO not available for 'raspberry' device!"
        LOG.info(result)



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import httplib2
import json

from iotronic_lightningrod.lightningrod import board

from oslo_log import log as logging
LOG = logging.getLogger(__name__)


def getBoardID():
    return board.uuid


def getLocation():
    return board.location


def getBoardGpio():
    return board.device.gpio


def sendRequest(url, action, headers=None, body=None, verbose=False):
    """Generic REST client for plugin users.

    :param url: resource URI
    :param action: POST, GET, PUT, etc
    :param headers: request header
    :param data: request body
    :param verbose: flag to enable/disable verbose output
    :return:

    """
    try:

        http = httplib2.Http()
        headers = headers
        response, send = http.request(url, action, headers=headers, body=body)

        if verbose:
            req = json.loads(send)
            LOG.info("Request:\n" + str(req) + "\n")
            LOG.info("Request response:\n" + str(response) + "\n")
            # LOG.info("\nREST REQUEST: HTTP " + str(response['status'])
            #         + " - success = " + str(req['success'])
            #         + " - " + str(req['result']['records']))

    except Exception as err:
        req = json.loads(send)
        LOG.error("Request Error:\n" + str(req) + "\n")
        LOG.error("sendRequest error: " + str(err))

    return response, send



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = "Nicola Peditto <n.peditto@gmail.com>"

import importlib as imp
import inspect
import os
import subprocess
import sys
import threading
import time

from datetime import datetime

from iotronic_lightningrod.config import package_path
from iotronic_lightningrod.lightningrod import RPC_devices
from iotronic_lightningrod.lightningrod import SESSION
from iotronic_lightningrod.modules import Module
from iotronic_lightningrod.modules import utils
import iotronic_lightningrod.wampmessage as WM


from oslo_log import log as logging
LOG = logging.getLogger(__name__)


class DeviceManager(Module.Module):

    def __init__(self, board, session):

        # Module declaration
        super(DeviceManager, self).__init__("DeviceManager", board)

        self.device_session = session

        device_type = board.type

        path = package_path + "/devices/" + device_type + ".py"

        if os.path.exists(path):

            device_module = imp.import_module(
                "iotronic_lightningrod.devices." + device_type
            )

            LOG.info(" - Device '" + device_type + "' module imported!")

            device = device_module.System()

            dev_meth_list = inspect.getmembers(
                device,
                predicate=inspect.ismethod
            )

            RPC_devices[device_type] = dev_meth_list

            self._deviceWampRegister(dev_meth_list, board)

            board.device = device

        else:
            LOG.warning("Device '" + device_type + "' not supported!")

    def finalize(self):
        pass

    def restore(self):
        pass

    def _deviceWampRegister(self, dev_meth_list, board):

        LOG.info(" - " + str(board.type).capitalize()
                 + " device registering RPCs:")

        for meth in dev_meth_list:

            if (meth[0] != "__init__") & (meth[0] != "finalize"):
                # LOG.info(" - " + str(meth[0]))
                # rpc_addr = u'iotronic.' + board.uuid + '.' + meth[0]
                rpc_addr = u'iotronic.' + str(board.session_id) + '.' + \
                           board.uuid + '.' + meth[0]

                # LOG.debug(" --> " + str(rpc_addr))
                SESSION.register(meth[1], rpc_addr)

                LOG.info("   --> " + str(meth[0]) + " registered!")

    async def DevicePing(self):
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        message = datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def DeviceReboot(self):
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        def delayBoardReboot():
            time.sleep(3)
            subprocess.call("reboot", shell=True)

        threading.Thread(target=delayBoardReboot).start()

        message = "Rebooting board in few seconds @" + \
                  str(datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'))
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def DeviceRestartLR(self):
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        def delayLRrestarting():
            time.sleep(2)
            python = sys.executable
            os.execl(python, python, *sys.argv)

        threading.Thread(target=delayLRrestarting).start()

        message = "Restarting LR in 5 seconds (" + \
                  datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f') + ")..."
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def DeviceHostname(self):
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        command = "hostname"

        out = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE
        )

        output = out.communicate()[0].decode('utf-8').strip()

        message = str(output) + "@" + \
            str(datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'))
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()

    async def DeviceNetConfig(self):
        rpc_name = utils.getFuncName()
        LOG.info("RPC " + rpc_name + " CALLED")

        command = "ifconfig"

        out = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE
        )

        output = out.communicate()[0].decode('utf-8').strip()

        message = str(output)
        w_msg = WM.WampSuccess(message)

        return w_msg.serialize()



# Copyright 2017 MDSLAB - University of Messina
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import asyncio
from iotronic_lightningrod.modules import Module
import iotronic_lightningrod.wampmessage as WM
from oslo_log import log as logging
import subprocess
import time
from urllib.parse import urlparse

LOG = logging.getLogger(__name__)

Port = []

interface = ""


class NetworkManager(Module.Module):

    def __init__(self, board, session):

        super(NetworkManager, self).__init__("NetworkManager", board)
        self.url_ip = urlparse(board.wamp_config["url"])[1].split(':')[0]
        self.wagent_url = "ws://" + self.url_ip + ":8080"

    def finalize(self):
        pass

    def restore(self):
        pass

    async def Create_VIF(self, r_tcp_port):

        LOG.info("Creation of the VIF ")

        inter_num = 30000 - int(r_tcp_port)
        global Port
        Port.insert(0, inter_num)

        LOG.debug("Creation of the VIF iotronic" + str(r_tcp_port))

        try:

            p1 = subprocess.Popen('socat -d -d TCP-L:' + str(inter_num) +
                                  ',bind=localhost,reuseaddr,forever,'
                                  'interval=10 TUN,tun-type=tap,'
                                  'tun-name=iotronic'
                                  + str(inter_num) + ',up  ',
                                  shell=True,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)

            p1 = subprocess.Popen('wstun -r' + str(r_tcp_port) + ':localhost:'
                                  + str(inter_num) + ' '
                                  + str(self.wagent_url),
                                  shell=True,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)

            LOG.debug('Creation of the VIF succeded: iotronic')

            global interface
            interface = 'iotronic' + str(inter_num)
            message = 'WS tun and SOCAT created'
            w_msg = WM.WampSuccess(message)

        except Exception:

            LOG.error('Error while creating the virtual interface')
            message = 'Error while the creation'
            w_msg = WM.WampError(message)

        return w_msg.serialize()

    async def Configure_VIF(self, port, cidr):

        LOG.info("Configuration of the VIF")

        try:
            LOG.debug("Configuration of the VIF " + interface)

            p3 = subprocess.Popen("ip link set dev " + interface + " address "
                                  + str(port['MAC_add']),
                                  shell=True,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)

            time.sleep(1)

            p5 = subprocess.Popen("ip address add " + str(port['ip']) + "/"
                                  + str(cidr) + " dev " + interface,
                                  shell=True,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT)

            message = 'IP address assigned'
            w_msg = WM.WampSuccess(message)

            LOG.info("Configuration succeded")

        except Exception as e:

            LOG.error(str(e))
            message = 'Error while the configuration'
            w_msg = WM.WampError(message)

        return w_msg.serialize()

    async def Remove_VIF(self, VIF_name):

        LOG.info("Removing a VIF from the board")

        try:
            LOG.info("Removing VIF")
            inter_num = 30000 - int(VIF_name[8:])
            LOG.debug("Removing VIF : iotronic" + str(inter_num))

            p1 = subprocess.\
                Popen("kill $(ps aux | grep -e '-r'"
                      + str(VIF_name[8:]) + " | awk '{print $2}')",
                      shell=True,
                      stdout=subprocess.PIPE,
                      stderr=subprocess.STDOUT)

            global Port
            Port.remove(inter_num)

            message = 'VIF removed'
            w_msg = WM.WampSuccess(message)

            LOG.info("VIF removed")
        except Exception as e:

            LOG.error(str(e))
            message = 'Error while removing the VIF'
            w_msg = WM.WampError(message)

        return w_msg.serialize()
